# Copyright (c) Acconeer AB, 2022
# All rights reserved

from acconeer.exptool import a121
from acconeer.exptool.a121.algo.distance import DetectorConfig


def get_default_detector_config() -> DetectorConfig:
    return DetectorConfig(
        end_m=1.0,
        max_profile=a121.Profile.PROFILE_1,
        threshold_sensitivity=0.7,
        signal_quality=25.0,
        update_rate=20.0,
    )



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from abc import abstractmethod


class DeviceFlasherBase:
    @staticmethod
    @abstractmethod
    def flash(device_port, device_name, image_path, progress_callback=None):
        pass



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import time
from typing import Any, Dict, Iterator, List, Optional, Union, cast

from acconeer.exptool.a121 import (
    ClientBase,
    ClientError,
    ClientInfo,
    Metadata,
    Record,
    Recorder,
    Result,
    SensorCalibration,
    SensorConfig,
    ServerInfo,
    SessionConfig,
)
from acconeer.exptool.a121._rate_calc import _RateCalculator, _RateStats


class _StopReplay(Exception):
    pass


class _ReplayingClient(ClientBase):
    _rate_stats_calc: Optional[_RateCalculator]

    def __init__(self, record: Record):
        self._record = record
        self._is_started: bool = False
        self._result_iterator: Optional[
            Union[Iterator[Result], Iterator[list[dict[int, Result]]]]
        ] = None
        self._origin_time: Optional[float] = None
        self._rate_stats_calc = None

    def _assert_connected(self) -> None:
        if not self.connected:
            raise ClientError("Client is not connected.")

    def _assert_session_setup(self) -> None:
        self._assert_connected()
        if not self.session_is_setup:
            raise ClientError("Session is not set up.")

    def _assert_session_started(self) -> None:
        self._assert_session_setup()
        if not self.session_is_started:
            raise ClientError("Session is not started.")

    def connect(self) -> None:
        pass

    def setup_session(
        self,
        config: Union[SensorConfig, SessionConfig],
        calibrations: Optional[dict[int, SensorCalibration]] = None,
    ) -> Union[Metadata, list[dict[int, Metadata]]]:
        if not self.connected:
            self.connect()

        if isinstance(config, SensorConfig):
            config = SessionConfig(config)

        if config != self._record.session_config:
            raise ValueError

        if self.session_config.extended:
            return self._record.extended_metadata
        else:
            return self._record.metadata

    def start_session(self, recorder: Optional[Recorder] = None) -> None:
        if recorder is not None:
            raise ValueError(f"{type(self).__name__} can not record")

        if self.session_config.extended:
            self._result_iterator = self._record.extended_results
        else:
            self._result_iterator = self._record.results

        self._is_started = True
        self._origin_time = None

        if self.session_config.extended:
            self._rate_stats_calc = _RateCalculator(
                self.session_config, self._record.extended_metadata
            )
        else:
            self._rate_stats_calc = _RateCalculator(self.session_config, self._record.metadata)

    def get_next(self) -> Union[Result, list[dict[int, Result]]]:
        if not self.session_is_setup:
            raise ClientError("Session is not set up.")

        assert self._result_iterator is not None

        try:
            result_ = next(self._result_iterator)
            result = cast(Union[Result, List[Dict[int, Result]]], result_)
        except StopIteration:
            raise _StopReplay

        if isinstance(result, Result):
            some_result = result
        else:
            some_result = next(iter(next(iter(result)).values()))

        now = time.monotonic() - some_result.tick_time

        if self._origin_time is None:
            self._origin_time = now

        delta = now - self._origin_time

        if delta < 0:
            time.sleep(-delta)

        assert self._rate_stats_calc is not None
        self._rate_stats_calc.update(result)

        return result

    def stop_session(self) -> Any:
        self._result_iterator = None
        self._is_started = False

    def disconnect(self) -> None:
        pass

    @property
    def connected(self) -> bool:
        return True

    @property
    def session_is_setup(self) -> bool:
        return True

    @property
    def session_is_started(self) -> bool:
        return self._is_started

    @property
    def server_info(self) -> ServerInfo:
        return self._record.server_info

    @property
    def client_info(self) -> ClientInfo:
        return self._record.client_info

    @property
    def session_config(self) -> SessionConfig:
        return self._record.session_config

    @property
    def extended_metadata(self) -> list[dict[int, Metadata]]:
        return self._record.extended_metadata

    @property
    def calibrations(self) -> dict[int, SensorCalibration]:
        return self._record.calibrations

    @property
    def calibrations_provided(self) -> dict[int, bool]:
        return self._record.calibrations_provided

    @property
    def _rate_stats(self) -> _RateStats:
        self._assert_session_started()
        assert self._rate_stats_calc is not None
        return self._rate_stats_calc.stats



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import os

from acconeer.exptool.setup.base import (
    PlatformInstall,
    RequireFileContentStep,
    ShellCommandStep,
    utils,
)


@PlatformInstall.register
class Linux(PlatformInstall):
    FT4222_UDEV_RULE_FILE = "/etc/udev/rules.d/50-ft4222.rules"
    FT4222_UDEV_RULE = (
        'SUBSYSTEM=="usb", ATTRS{idVendor}=="0403", ATTRS{idProduct}=="601c", MODE:="0666"\n'
    )

    XC120_UDEV_RULE_FILE = "/etc/udev/rules.d/50-xc120.rules"
    XC120_UDEV_RULE = (
        'SUBSYSTEM=="usb", ATTRS{idVendor}=="0483", ATTRS{idProduct}=="a41d", MODE:="0666"\n'
        'SUBSYSTEM=="usb", ATTRS{idVendor}=="0483", ATTRS{idProduct}=="a42c", MODE:="0666"\n'
        'SUBSYSTEM=="usb", ATTRS{idVendor}=="0483", ATTRS{idProduct}=="a42d", MODE:="0666"\n'
        'SUBSYSTEM=="usb", ATTRS{idVendor}=="0483", ATTRS{idProduct}=="a449", MODE:="0666"\n'
    )

    def __init__(self) -> None:
        super().__init__(
            utils.WithDescription(
                "> Setup up permissions needed for UART communication "
                + "by adding the current user to the 'dialout' group.",
                ShellCommandStep(f"sudo usermod -a -G dialout {os.environ.get('USER')}".split()),
            ),
            utils.WithDescription(
                "> Create an udev rule for SPI communication with XM112.",
                RequireFileContentStep(
                    self.FT4222_UDEV_RULE_FILE, self.FT4222_UDEV_RULE, sudo=True
                ),
            ),
            utils.WithDescription(
                "> Create an udev rule for USB communication with XC120.",
                RequireFileContentStep(self.XC120_UDEV_RULE_FILE, self.XC120_UDEV_RULE, sudo=True),
            ),
        )

    @classmethod
    def get_key(cls) -> str:
        return "Linux"



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from .config_enums import PRF, IdleState, Profile
from .sensor_config import SensorConfig
from .session_config import SessionConfig
from .subsweep_config import SubsweepConfig
from .validation_error import Criticality, ValidationError, ValidationResult, ValidationWarning



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import acconeer.exptool as et
from acconeer.exptool import a121


args = a121.ExampleArgumentParser().parse_args()
et.utils.config_logging(args)

client = a121.Client(**a121.get_client_args(args))
client.connect()

session_config = a121.SessionConfig(
    [
        {
            1: a121.SensorConfig(
                sweeps_per_frame=10,
            )
        },
        {
            1: a121.SensorConfig(
                sweeps_per_frame=20,
            )
        },
    ],
    extended=True,
)

client.setup_session(session_config)
client.start_session()

for i in range(3):
    extended_result = client.get_next()

    print(f"\nExtended result {i + 1}:")
    print(extended_result)

client.stop_session()
client.disconnect()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import abc
import platform
import webbrowser
from typing import List, Optional

import pyperclip
import qtawesome as qta

from PySide6 import QtCore, QtGui
from PySide6.QtWidgets import (
    QComboBox,
    QFrame,
    QHBoxLayout,
    QLabel,
    QMessageBox,
    QPushButton,
    QWidget,
)

from acconeer.exptool.app.new._enums import ConnectionInterface
from acconeer.exptool.app.new._exceptions import HandledException
from acconeer.exptool.app.new.app_model import AppModel
from acconeer.exptool.utils import CommDevice  # type: ignore[import]


BUTTON_ICON_COLOR = "#0081db"


class ExceptionWidget(QMessageBox):
    def __init__(
        self,
        parent: QWidget,
        *,
        exc: Exception,
        traceback_str: Optional[str] = None,
        title: str = "Error",
    ) -> None:
        super().__init__(parent)

        self.setIcon(QMessageBox.Warning)
        self.setStandardButtons(QMessageBox.Ok)

        self.setWindowTitle(title)
        self.setText(str(exc))

        try:
            raise exc
        except HandledException:
            pass
        except Exception:
            self.setInformativeText("<b>Unhandled error - please file a bug</b>")

        if traceback_str:
            self.setDetailedText(traceback_str)
            copy_button = QPushButton(self)
            copy_button.setText("Copy details")
            self.addButton(copy_button, QMessageBox.ButtonRole.ActionRole)
            copy_button.clicked.disconnect()
            copy_button.clicked.connect(self._on_copy_clicked)

    def resizeEvent(self, event: QtGui.QResizeEvent) -> None:
        super().resizeEvent(event)
        self.setFixedWidth(500)

    def _on_copy_clicked(self) -> None:
        detailed_text = self.detailedText()
        if detailed_text:
            pyperclip.copy(detailed_text)

    def closeEvent(self, event: QtGui.QCloseEvent) -> None:
        self.close()


class VerticalSeparator(QWidget):
    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent)

        self.setLayout(QHBoxLayout(self))
        self.layout().setContentsMargins(5, 5, 5, 5)

        frame = QFrame(self)
        frame.setFrameShape(QFrame.VLine)
        self.layout().addWidget(frame)


class CommDeviceComboBox(QComboBox):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)
        self.app_model = app_model
        app_model.sig_notify.connect(self._on_app_model_update)
        self.currentTextChanged.connect(self._on_change)
        self.setMinimumWidth(175)

    def _on_app_model_update(self, app_model: AppModel) -> None:
        devices = self._get_available_devices()

        view_ports = [self.itemData(i) for i in range(self.count())]

        with QtCore.QSignalBlocker(self):
            if devices != view_ports:
                self.clear()
                for device in devices:
                    self.addItem(device.display_name(), device)

            current_device = self._get_current_device()

            # QComboBox.findData can't be used since since it doesn't
            # always match different but content wise identical objects.
            index = -1
            for i in range(self.count()):
                if self.itemData(i) == current_device:
                    index = i
                    break
            self.setCurrentIndex(index)

            self.setEnabled(self.count() > 0)
            if self.count() == 0:
                self.addItem("No device available")

    @abc.abstractmethod
    def _on_change(self) -> None:
        pass

    @abc.abstractmethod
    def _get_available_devices(self) -> List[CommDevice]:
        pass

    @abc.abstractmethod
    def _get_current_device(self) -> Optional[CommDevice]:
        pass


class SerialPortComboBox(CommDeviceComboBox):
    def _on_change(self) -> None:
        self.app_model.set_serial_connection_device(self.currentData())

    def _get_available_devices(self) -> List[CommDevice]:
        return self.app_model.available_serial_devices

    def _get_current_device(self) -> Optional[CommDevice]:
        return self.app_model.serial_connection_device


class USBDeviceComboBox(CommDeviceComboBox):
    def _on_change(self) -> None:
        self.app_model.set_usb_connection_device(self.currentData())

    def _get_available_devices(self) -> List[CommDevice]:
        return self.app_model.available_usb_devices

    def _get_current_device(self) -> Optional[CommDevice]:
        return self.app_model.usb_connection_device


class HintObject:
    def __init__(self, warning: str, tooltip: str, how_to_fix_url: str) -> None:
        self.warning = warning
        self.tooltip = tooltip
        self.how_to_fix_url = how_to_fix_url

    @staticmethod
    @abc.abstractmethod
    def _should_show(app_model: AppModel) -> bool:
        pass


class UserHintWidget(QWidget):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)
        app_model.sig_notify.connect(self._on_app_model_update)

        self._hints: List[HintObject] = []
        self._how_to_fix_url: Optional[str] = None

        self.setLayout(QHBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)

        self.icon = qta.IconWidget()
        self.icon.setHidden(True)
        self.icon.setIcon(qta.icon("fa.warning", color="#ff9e00"))
        self.layout().addWidget(self.icon)

        self.label = QLabel(self)
        self.label.setHidden(True)
        self.layout().addWidget(self.label)

        self.button = QPushButton(self)
        self.button.setIcon(qta.icon("fa5s.external-link-alt", color=BUTTON_ICON_COLOR))
        self.button.setText("How to fix")
        self.button.clicked.connect(self._on_click)
        self.button.setHidden(True)
        self.layout().addWidget(self.button)

    def add_hint(self, hint: HintObject) -> None:
        self._hints.append(hint)

    def _on_app_model_update(self, app_model: AppModel) -> None:
        for hint in self._hints:
            if hint._should_show(app_model):
                self.label.setText(hint.warning)
                self.label.setToolTip(hint.tooltip)
                self.icon.setToolTip(hint.tooltip)
                self._how_to_fix_url = hint.how_to_fix_url
                self.icon.setHidden(False)
                self.label.setHidden(False)
                if self._how_to_fix_url is not None:
                    self.button.setHidden(False)
                return

        self.icon.setHidden(True)
        self.label.setHidden(True)
        self.button.setHidden(True)

    def _on_click(self) -> None:
        if self._how_to_fix_url is not None:
            webbrowser.open_new_tab(self._how_to_fix_url)


class ConnectionHint(HintObject):
    def __init__(self) -> None:
        super().__init__(
            "Stability warning",
            "You may experience stability issues due to windows serial port driver",
            r"https://docs.acconeer.com/en/latest/evk_setup/xc120_xe121.html",
        )

    @staticmethod
    def _should_show(app_model: AppModel) -> bool:
        if platform.system().lower() != "windows":
            return False

        if (
            app_model.serial_connection_device is not None
            and app_model.connection_interface == ConnectionInterface.SERIAL
        ):
            if app_model.serial_connection_device.name is not None:
                if app_model.serial_connection_device.unflashed:
                    return False
                return "xc120" in app_model.serial_connection_device.name.lower()

        return False


class UnflashedDeviceHint(HintObject):
    def __init__(self) -> None:
        super().__init__(
            "Unflashed device",
            "The device needs to be flashed with exploration server firmware",
            r"https://docs.acconeer.com/en/latest/evk_setup/xc120_xe121.html",
        )

    @staticmethod
    def _should_show(app_model: AppModel) -> bool:
        if app_model.connection_interface not in [
            ConnectionInterface.SERIAL,
            ConnectionInterface.USB,
        ]:
            return False

        if (
            app_model.serial_connection_device is not None
            and app_model.connection_interface == ConnectionInterface.SERIAL
            and app_model.serial_connection_device.unflashed
        ):
            return True

        if (
            app_model.usb_connection_device is not None
            and app_model.connection_interface == ConnectionInterface.USB
            and app_model.usb_connection_device.unflashed
        ):
            return True

        return False


class InaccessibleDeviceHint(HintObject):
    def __init__(self) -> None:
        super().__init__(
            "Device permissions",
            "The USB device permissions needs to be setup, "
            "update USB permissions or use Serial Port",
            r"https://docs.acconeer.com/en/latest/exploration_tool/"
            "installation_and_setup.html#linux-setup",
        )

    @staticmethod
    def _should_show(app_model: AppModel) -> bool:
        if (
            app_model.usb_connection_device is not None
            and app_model.connection_interface == ConnectionInterface.USB
            and not app_model.usb_connection_device.accessible
        ):
            return True

        return False


class HintWidget(QWidget):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)
        self.setLayout(QHBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)

        hint_widget = UserHintWidget(app_model, self)
        self.layout().addWidget(hint_widget)

        # Prioritized hint order:
        # The first will have priority over the second
        # The second will have priority over the third...
        hint_widget.add_hint(InaccessibleDeviceHint())
        hint_widget.add_hint(UnflashedDeviceHint())
        hint_widget.add_hint(ConnectionHint())



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from typing import Optional

from acconeer.exptool.a121._core.entities import PRF, SensorCalibration, SessionConfig
from acconeer.exptool.a121._core.utils import iterate_extended_structure_values

from ._latest import ExplorationProtocol, ExplorationProtocolError


class ExplorationProtocol_No_15_6MHz_PRF(ExplorationProtocol):
    @classmethod
    def setup_command(
        cls,
        session_config: SessionConfig,
        calibrations: Optional[dict[int, SensorCalibration]] = None,
    ) -> bytes:
        for sensor_config in iterate_extended_structure_values(session_config.groups):
            if any(ssc.prf == PRF.PRF_15_6_MHz for ssc in sensor_config.subsweeps):
                raise ExplorationProtocolError(
                    "Connected Exploration server does not support 15.6MHz PRF."
                )

        return super().setup_command(session_config, calibrations)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import numpy as np

import pyqtgraph as pg

from acconeer.exptool import utils


class PGUpdater:
    def __init__(self, sensor_config, processing_config, session_info):
        self.session_info = session_info
        self.smooth_max = utils.SmoothMax(sensor_config.update_rate)

    def setup(self, win):
        num_depths = self.session_info["bin_count"]
        start = self.session_info["range_start_m"]
        length = self.session_info["range_length_m"]
        end = start + length

        xs = np.linspace(start, end, num_depths * 2 + 1)[1::2]
        bin_width = 0.8 * length / num_depths

        self.plot = win.addPlot()
        self.plot.setMenuEnabled(False)
        self.plot.setMouseEnabled(x=False, y=False)
        self.plot.hideButtons()
        self.plot.showGrid(x=True, y=True)
        self.plot.setLabel("bottom", "Depth (m)")
        self.plot.setLabel("left", "Amplitude")
        self.plot.setXRange(start, end)
        self.plot.setYRange(0, 1)

        self.bar_graph = pg.BarGraphItem(
            x=xs,
            height=np.zeros_like(xs),
            width=bin_width,
            brush=pg.mkBrush(utils.color_cycler()),
        )

        self.plot.addItem(self.bar_graph)

    def update(self, data):
        self.bar_graph.setOpts(height=data)
        self.plot.setYRange(0, self.smooth_max.update(data))



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import logging
from typing import Any, Callable, Optional, TypeVar

import packaging.version

from acconeer.exptool import a121
from acconeer.exptool.app.new._enums import ConnectionState, PluginState
from acconeer.exptool.app.new._exceptions import HandledException

from ._backend_plugin import BackendPlugin
from ._message import ConnectionStateMessage, GeneralMessage, Message, PluginStateMessage


log = logging.getLogger(__name__)


T = TypeVar("T")
MessageHandler = Callable[[Message], None]
BackendPluginFactory = Callable[[MessageHandler, str], BackendPlugin]


def is_task(func: T) -> T:
    setattr(func, "is_task", True)
    return func


class Model:
    backend_plugin: Optional[BackendPlugin]
    client: Optional[a121.Client]

    def __init__(self, task_callback: Callable[[Message], None]) -> None:
        self.backend_plugin = None
        self.client = None
        self.task_callback = task_callback

    def idle(self) -> bool:
        if self.backend_plugin is None:
            return False

        return self.backend_plugin.idle()

    def execute_task(self, name: str, kwargs: dict[str, Any], plugin: bool) -> None:
        if plugin:
            if self.backend_plugin is None:
                raise RuntimeError

            obj: Any = self.backend_plugin
        else:
            obj = self

        try:
            method = getattr(obj, name)
            if not getattr(method, "is_task"):
                raise Exception
        except Exception:
            raise RuntimeError(f"'{name}' is not a task")

        method(**kwargs)

    @is_task
    def connect_client(self, client_info: a121.ClientInfo) -> None:
        if self.client is not None:
            raise RuntimeError(
                "Model already has a Client. The current Client needs to be disconnected first."
            )

        self.client = a121.Client(
            ip_address=client_info.ip_address,
            serial_port=client_info.serial_port,
            usb_device=client_info.usb_device,
            override_baudrate=client_info.override_baudrate,
            mock=client_info.mock,
        )

        try:
            self.client.connect()
        except Exception as exc:
            self.client = None
            self.task_callback(ConnectionStateMessage(state=ConnectionState.DISCONNECTED))

            msg = "Failed to connect"

            try:
                msg += f":\n{exc}"
            except Exception:
                pass

            raise HandledException(msg)

        if self.backend_plugin is not None:
            self.backend_plugin.attach_client(client=self.client)

        ver = packaging.version.Version(a121.SDK_VERSION)
        if self.client.server_info.parsed_rss_version > ver:
            connection_warning = "New server version - please upgrade client"
        elif self.client.server_info.parsed_rss_version < ver:
            connection_warning = "Old server version - please upgrade server"
        else:
            connection_warning = None

        self.task_callback(
            ConnectionStateMessage(state=ConnectionState.CONNECTED, warning=connection_warning)
        )
        self.task_callback(GeneralMessage(name="server_info", data=self.client.server_info))

    @is_task
    def disconnect_client(self) -> None:
        if self.client is None:
            raise RuntimeError("Backend has no client to disconnect.")

        if self.backend_plugin is not None:
            self.backend_plugin.detach_client()

        try:
            self.client.disconnect()
        except Exception:
            pass

        self.client = None

        self.task_callback(ConnectionStateMessage(state=ConnectionState.DISCONNECTED))

    @is_task
    def load_plugin(self, *, plugin_factory: BackendPluginFactory, key: str) -> None:
        if self.backend_plugin is not None:
            self.unload_plugin(send_callback=False)

        self.backend_plugin = plugin_factory(self.task_callback, key)
        log.info(f"{plugin_factory.__name__} was loaded.")

        if self.client is not None and self.client.connected:
            self.backend_plugin.attach_client(client=self.client)
            log.debug(f"{plugin_factory.__name__} was attached a Client")

        self.task_callback(PluginStateMessage(state=PluginState.LOADED_IDLE))

    @is_task
    def unload_plugin(self, send_callback: bool = True) -> None:
        if self.backend_plugin is None:
            return

        self.backend_plugin.teardown()
        self.backend_plugin = None
        log.debug("Current BackendPlugin was torn down")

        if send_callback:
            self.task_callback(PluginStateMessage(state=PluginState.UNLOADED))



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import acconeer.exptool as et
from acconeer.exptool import a121


args = a121.ExampleArgumentParser().parse_args()
et.utils.config_logging(args)

"""
First session - Setup session without calibration
                Sensor will calibrate during setup
"""

client = a121.Client(**a121.get_client_args(args))
client.connect()

sensor_config = a121.SensorConfig()
client.setup_session(sensor_config)

calibrations = client.calibrations

"""
client.calibrations_provided will be equal to {1: False} since calibration was not provided
"""
print(f"Setup without provided calibration {client.calibrations_provided}")

client.disconnect()

"""
Second session - Setup session with calibration
                 Sensor will reuse calibration and not calibrate during setup
"""

client = a121.Client(**a121.get_client_args(args))
client.connect()

sensor_config = a121.SensorConfig()
client.setup_session(sensor_config, calibrations)

"""
client.calibrations_provided will be equal to {1: True} since calibration was provided
"""
print(f"Setup with provided calibration {client.calibrations_provided}")

client.disconnect()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import enum
import importlib.resources
import operator
from functools import partial, reduce

import attr
import yaml

from acconeer.exptool.a111 import _configs
from acconeer.exptool.a111._modes import Mode, get_mode

from . import data


BYTEORDER = "little"
BO = BYTEORDER


class Category(enum.Enum):
    GENERAL = "general"
    CONFIG = "config"
    SESSION_INFO = "metadata"
    DATA_INFO = "result_info"
    RESULT = "result"


class DataType(enum.Enum):
    BITSET = "bitmask"
    ENUM = "enum"
    BOOL = "boolean"
    UINT16 = "uint16"
    UINT32 = "uint32"
    INT32 = "int32"


@attr.s
class Register:
    full_name = attr.ib()
    stripped_name = attr.ib()
    addr = attr.ib()
    modes = attr.ib()
    readable = attr.ib()
    writable = attr.ib()
    category = attr.ib()
    data_type = attr.ib()

    float_scale = attr.ib(default=None)  # Applicable iff float, if not None then decode as float

    enum = attr.ib(default=None)
    bitset_flags = attr.ib(default=None)
    bitset_masks = attr.ib(default=None)

    def encode(self, value):
        if self.data_type == DataType.BITSET:
            try:
                x = int(self.bitset_flags(value))
            except ValueError:
                if isinstance(value, str):
                    x = self.bitset_flags[value.upper()].value
                else:
                    xs = [self.bitset_flags[v.upper()].value for v in value] + [0]
                    x = reduce(operator.or_, xs)

            return x.to_bytes(4, BO)

        if self.data_type == DataType.ENUM:
            if isinstance(value, enum.Enum) and not isinstance(value, self.enum):
                value = ENUM_REMAP.get(value, value.name)

            if isinstance(value, str):
                value = self.enum[value.upper()]
            elif isinstance(value, int):
                value = self.enum(value)

            if not isinstance(value, self.enum):
                raise ValueError

            return value.value.to_bytes(4, BO)

        if self.data_type == DataType.BOOL:
            return int(bool(value)).to_bytes(4, BO)

        if self.float_scale is not None:
            value = round(value * self.float_scale)

        value = int(value)
        signed = self.data_type == DataType.INT32

        if not signed and value < 0:
            raise ValueError

        return value.to_bytes(4, BO, signed=signed)

    def decode(self, value):
        signed = self.data_type == DataType.INT32
        value = int.from_bytes(value, BO, signed=signed)

        if self.data_type == DataType.BITSET:
            return self.bitset_flags(value)

        if self.data_type == DataType.ENUM:
            return self.enum(value)

        if self.data_type == DataType.BOOL:
            return bool(value)

        if self.float_scale is not None:
            return float(value) / self.float_scale

        return value


PREFIX_TO_MODE_MAP = {
    "pb": Mode.POWER_BINS,
    "env": Mode.ENVELOPE,
    "iq": Mode.IQ,
    "sp": Mode.SPARSE,
    "sparse": Mode.SPARSE,
    "peak": None,
    "obst": None,
    "pres": None,
}

# Is tested for completeness
# Order can be important
CONFIG_TO_STRIPPED_REG_NAME_MAP = {
    "mode": None,  # "mode_selection",  # Explicitly set in client
    "profile": "profile_selection",
    "sampling_mode": "sampling_mode",
    "repetition_mode": "repetition_mode",
    "update_rate": "update_rate",
    "sweep_rate": "req_sweep_rate",
    "range_start": "range_start",
    "range_length": "range_length",
    "gain": "gain",
    "downsampling_factor": "downsampling_factor",
    "hw_accelerated_average_samples": "hw_acc_average_samples",
    "bin_count": "req_bin_count",
    "running_average_factor": "run_factor",
    "sweeps_per_frame": "sweeps_per_frame",
    "noise_level_normalization": "noise_level_normalization",
    "maximize_signal_attenuation": "maximize_signal_attenuation",
    "tx_disable": "tx_disable",
    "asynchronous_measurement": "asynchronous_measurement",
    "power_save_mode": "sensor_power_mode",
    "mur": "mur",
    "_depth_lowpass_cutoff_ratio_override": "depth_lpf_ratio_override",
    "_depth_lowpass_cutoff_ratio_value": "depth_lpf_ratio_value",
    "depth_lowpass_cutoff_ratio": None,
    "range_end": None,
    "range_interval": None,
    "sensor": None,
}

STRIPPED_NAME_TO_INFO_REMAP = {
    "output_buffer_length": None,
    "start": "range_start_m",
    "length": "range_length_m",
    "step_length": "step_length_m",
    "step_length": "step_length_m",
    "depth_lpf_ratio_used": "depth_lowpass_cutoff_ratio",
}

ENUM_REMAP = {
    _configs.BaseServiceConfig.RepetitionMode.SENSOR_DRIVEN: "streaming",
    _configs.BaseServiceConfig.RepetitionMode.HOST_DRIVEN: "on_demand",
}

REGISTERS = None


def _match_reg_by_addr(addr, reg):
    return reg.addr == addr


def _match_reg_by_name(name, reg):
    return name in (reg.full_name, reg.stripped_name)


def get_reg(value, mode=None):
    if isinstance(value, Register):
        return value
    elif isinstance(value, int):
        match_fun = partial(_match_reg_by_addr, value)
    elif isinstance(value, str):
        match_fun = partial(_match_reg_by_name, value)
    else:
        raise ValueError

    mode = get_mode(mode)
    matches = []

    for reg in REGISTERS:
        if match_fun(reg):
            if mode is None or reg.modes is None or mode in reg.modes:
                matches.append(reg)

    if len(matches) < 1:
        raise ValueError("unknown reg: {}".format(value))

    if len(matches) > 1:
        raise ValueError("ambiguous reg: {}".format(value))

    return matches[0]


def get_reg_addr(name_or_reg_or_addr, mode=None):
    if isinstance(name_or_reg_or_addr, int):
        addr = name_or_reg_or_addr
        return addr

    name_or_reg = name_or_reg_or_addr
    return get_reg(name_or_reg, mode).addr


def get_regs_for_mode(mode):
    if mode is None:
        raise ValueError

    mode = get_mode(mode)
    return [reg for reg in REGISTERS if reg.modes is None or mode in reg.modes]


def get_regs_for_mode_in_category(category, mode):
    return [reg for reg in get_regs_for_mode(mode) if reg.category == Category(category)]


get_session_info_regs = partial(get_regs_for_mode_in_category, Category.SESSION_INFO)
get_data_info_regs = partial(get_regs_for_mode_in_category, Category.DATA_INFO)


def get_config_key_to_reg_map(mode):  # {config_key: reg}
    mode = get_mode(mode)
    config_cls = _configs.MODE_TO_CONFIG_CLASS_MAP[mode]

    m = {}
    for config_key, reg_name in CONFIG_TO_STRIPPED_REG_NAME_MAP.items():
        if reg_name is None:
            continue

        try:
            param = getattr(config_cls, config_key)
        except AttributeError:
            continue

        if param.is_dummy:
            continue

        m[config_key] = get_reg(reg_name, mode)

    return m


def load_yaml():
    global REGISTERS

    if REGISTERS is not None:
        return

    with importlib.resources.open_text(data, "regmap.yaml") as stream:
        raw_regs = yaml.safe_load(stream)

    REGISTERS = []

    for raw_name, raw_reg in raw_regs.items():
        raw_modes = raw_reg.get("modes", None)
        if raw_modes is None or raw_modes == "None":
            modes = None
        elif isinstance(raw_modes, str):
            try:
                modes = [get_mode(raw_modes)]
            except ValueError:
                continue
        else:  # assumed to be a list
            modes = []
            for m in raw_modes:
                try:
                    modes.append(get_mode(m))
                except ValueError:
                    pass

            if len(modes) == 0:
                continue

        full_name = raw_name.strip().lower()

        try:
            prefix, stripped_name = full_name.split("_", 1)
        except ValueError:
            stripped_name = full_name
        else:
            if prefix in PREFIX_TO_MODE_MAP.keys():
                mode = PREFIX_TO_MODE_MAP[prefix]

                if mode is None:
                    continue

                assert PREFIX_TO_MODE_MAP[prefix] == modes[0] and len(modes) == 1
            else:
                stripped_name = full_name

        addr = raw_reg["address"]
        assert type(addr) == int

        readable, writable = [c in raw_reg.get("access", "rw").strip().lower() for c in "rw"]
        category = Category(raw_reg["category"].strip().lower())
        data_type = DataType(raw_reg["type"].strip().lower())

        reg = Register(
            full_name=full_name,
            stripped_name=stripped_name,
            addr=addr,
            modes=modes,
            readable=readable,
            writable=writable,
            category=category,
            data_type=data_type,
        )

        try:
            reg.float_scale = float(raw_reg["scale"])
        except (KeyError, ValueError):
            pass
        else:
            assert reg.data_type == DataType.INT32

        if data_type == DataType.ENUM:
            enum_values = {str(k).upper(): int(d["value"]) for k, d in raw_reg["values"].items()}
            reg.enum = enum.IntEnum(full_name + "_enum", enum_values)

        if data_type == DataType.BITSET:
            flags = {}
            masks = {}

            for k, v in [(str(k).upper(), int(d["value"])) for k, d in raw_reg["bits"].items()]:
                if v == 0:
                    continue

                if v & (v - 1) == 0:  # is power of 2
                    flags[k] = v
                else:
                    masks[k] = v

            reg.bitset_flags = enum.IntFlag(full_name + "_bitset_flags", flags)
            reg.bitset_masks = enum.IntEnum(full_name + "_bitset_masks", masks)

        REGISTERS.append(reg)


load_yaml()

STATUS_REG = get_reg("status")
STATUS_FLAGS = STATUS_REG.bitset_flags
STATUS_MASKS = STATUS_REG.bitset_masks



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import time
from typing import Any, Optional, Union

import numpy as np
import numpy.typing as npt

from acconeer.exptool import a121
from acconeer.exptool.a121._core.entities import (
    INT_16_COMPLEX,
    ClientInfo,
    Metadata,
    Profile,
    Result,
    ResultContext,
    SensorCalibration,
    SensorConfig,
    SensorInfo,
    ServerInfo,
    SessionConfig,
    SubsweepConfig,
)
from acconeer.exptool.a121._core.mediators import ClientError, Recorder
from acconeer.exptool.a121._core.utils import unextend
from acconeer.exptool.a121._perf_calc import _SessionPerformanceCalc

from .common_client import CommonClient
from .utils import get_calibrations_provided


class MockClient(CommonClient):

    TICKS_PER_SECOND = 1000000
    CALIBRATION_TEMPERATURE = 25
    BASE_STEP_LENGTH_M = 0.0025
    SENSOR_COUNT = 5
    SENSOR_OBJECTS = {
        1: {"distance_mm": 500, "peak_amplitude": 10000, "phase": 0.5},
        2: {"distance_mm": 1000, "peak_amplitude": 8000, "phase": 1.0},
        3: {"distance_mm": 1500, "peak_amplitude": 6000, "phase": 1.5},
        4: {"distance_mm": 2000, "peak_amplitude": 4000, "phase": 2.0},
        5: {"distance_mm": 2500, "peak_amplitude": 2000, "phase": 2.5},
    }
    FWHM = {
        Profile.PROFILE_1: 16,
        Profile.PROFILE_2: 30,
        Profile.PROFILE_3: 54,
        Profile.PROFILE_4: 76,
        Profile.PROFILE_5: 132,
    }
    DIRECT_LEAKAGE_PHASE = 0.0
    DIRECT_LEAKAGE_AMPLITUDE = 2000
    NOISE_AMPLITUDE = 20
    MOCK_SERVER_INFO = ServerInfo(
        rss_version=f"a121-v{a121.SDK_VERSION}",
        sensor_count=SENSOR_COUNT,
        ticks_per_second=TICKS_PER_SECOND,
        hardware_name="Mock-system",
        sensor_infos={
            1: SensorInfo(connected=True, serial="SN1"),
            2: SensorInfo(connected=True, serial="SN2"),
            3: SensorInfo(connected=True, serial="SN3"),
            4: SensorInfo(connected=True, serial="SN4"),
            5: SensorInfo(connected=True, serial="SN5"),
        },
    )
    MIN_MOCK_UPDATE_RATE_HZ = 1.0
    MAX_MOCK_UPDATE_RATE_HZ = 100.0

    _client_info: ClientInfo
    _connected: bool
    _start_time: float
    _mock_update_rate: float
    _mock_next_data_time: float

    def __init__(self, client_info: ClientInfo) -> None:
        self._start_time = time.monotonic()
        self._connected = False
        self._mock_update_rate = self.MAX_MOCK_UPDATE_RATE_HZ
        self._mock_next_data_time = 0.0
        super().__init__(client_info)

    @classmethod
    def _sensor_config_to_metadata(
        cls, sensor_config: SensorConfig, update_rate: Optional[float]
    ) -> Metadata:
        subsweep_data_length = []
        subsweep_data_offset = []
        sweep_data_length = 0
        tick_period = 0 if update_rate is None else int(cls.TICKS_PER_SECOND / update_rate)
        for subsweep in sensor_config.subsweeps:
            subsweep_data_length.append(subsweep.num_points)
            subsweep_data_offset.append(sweep_data_length)
            sweep_data_length += subsweep.num_points
        return Metadata(
            frame_data_length=sensor_config._sweeps_per_frame * sweep_data_length,
            sweep_data_length=sweep_data_length,
            subsweep_data_length=np.array(subsweep_data_length),
            subsweep_data_offset=np.array(subsweep_data_offset),
            max_sweep_rate=100000,
            tick_period=tick_period,
            calibration_temperature=cls.CALIBRATION_TEMPERATURE,
            base_step_length_m=cls.BASE_STEP_LENGTH_M,
        )

    @classmethod
    def _session_config_to_metadata(cls, config: SessionConfig) -> list[dict[int, Metadata]]:
        metadata_list = []
        for group in config.groups:
            metadata_dict = {}
            for sensor_id, sensor_config in group.items():
                metadata_dict[sensor_id] = cls._sensor_config_to_metadata(
                    sensor_config, config.update_rate
                )
            metadata_list.append(metadata_dict)
        return metadata_list

    @classmethod
    def _get_mock_data(cls, sensor_id: int, subsweep: SubsweepConfig) -> npt.NDArray[np.complex_]:

        noise: npt.NDArray[np.complex_] = np.random.normal(
            0, cls.NOISE_AMPLITUDE, size=2 * subsweep.num_points
        ).view(np.complex_)

        if not subsweep.enable_tx:
            return noise

        object_distance = cls.SENSOR_OBJECTS[sensor_id]["distance_mm"] / (
            1000 * cls.BASE_STEP_LENGTH_M
        )
        peak_amplitude = cls.SENSOR_OBJECTS[sensor_id]["peak_amplitude"]
        phase = cls.SENSOR_OBJECTS[sensor_id]["phase"]
        points = subsweep.start_point + np.arange(subsweep.num_points) * subsweep.step_length
        std = cls.FWHM[subsweep.profile] / 2.355
        direct_leakage: npt.NDArray[np.complex_] = (
            np.exp(1j * cls.DIRECT_LEAKAGE_PHASE)
            * cls.DIRECT_LEAKAGE_AMPLITUDE
            * np.exp(-((points) ** 2) / (2 * std**2))
        )

        if subsweep.enable_loopback:
            return direct_leakage + noise

        signal: npt.NDArray[np.complex_] = (
            np.exp(1j * phase)
            * peak_amplitude
            * np.exp(-((points - object_distance) ** 2) / (2 * std**2))
        )

        return direct_leakage + signal + noise

    @classmethod
    def _sensor_config_to_frame(
        cls, sensor_id: int, sensor_config: SensorConfig, metadata: Metadata
    ) -> npt.NDArray:
        frame: npt.NDArray = np.ndarray(
            shape=(sensor_config._sweeps_per_frame, metadata.sweep_data_length),
            dtype=INT_16_COMPLEX,
        )
        for sweep in range(0, sensor_config._sweeps_per_frame):
            sweep_offset = 0
            for subsweep in sensor_config.subsweeps:
                subsweep_data = cls._get_mock_data(sensor_id, subsweep)
                for idx, point in enumerate(subsweep_data):
                    frame[sweep][sweep_offset + idx]["real"] = point.real
                    frame[sweep][sweep_offset + idx]["imag"] = point.imag
                sweep_offset += subsweep.num_points
        return frame

    def _sensor_config_to_result(self, sensor_id: int, sensor_config: SensorConfig) -> Result:
        metadata = self._sensor_config_to_metadata(sensor_config, update_rate=None)
        return Result(
            data_saturated=False,
            frame_delayed=False,
            calibration_needed=False,
            temperature=int(self.CALIBRATION_TEMPERATURE + np.random.normal(0, 2)),
            tick=int((time.monotonic() - self._start_time) * self.TICKS_PER_SECOND),
            frame=self._sensor_config_to_frame(sensor_id, sensor_config, metadata),
            context=ResultContext(ticks_per_second=self.TICKS_PER_SECOND, metadata=metadata),
        )

    def _session_config_to_result(self, config: SessionConfig) -> list[dict[int, Result]]:
        result_list = []
        for group in config.groups:
            result_dict = {}
            for sensor_id, sensor_config in group.items():
                result_dict[sensor_id] = self._sensor_config_to_result(sensor_id, sensor_config)
            result_list.append(result_dict)
        return result_list

    def connect(self) -> None:
        if not self._connected:
            self._connected = True
        else:
            raise ClientError("Client is already connected")

    def setup_session(
        self,
        config: Union[SensorConfig, SessionConfig],
        calibrations: Optional[dict[int, SensorCalibration]] = None,
    ) -> Union[Metadata, list[dict[int, Metadata]]]:
        if self.session_is_started:
            raise ClientError("Session is currently running, can't setup.")

        if isinstance(config, SensorConfig):
            config = SessionConfig(config)

        config.validate()

        self._calibrations_provided = get_calibrations_provided(config, calibrations)
        self._session_config = config
        self._metadata = self._session_config_to_metadata(config)

        self._sensor_calibrations = {}
        for group in config.groups:
            for sensor_id, sensor_config in group.items():
                self._sensor_calibrations[sensor_id] = SensorCalibration(
                    temperature=25,
                    data="mocked calibration",
                )
        pc = _SessionPerformanceCalc(config, self._metadata)
        self._mock_update_rate = pc.update_rate

        # Keep the mock update rate between 1Hz and 100Hz to both have a
        # responsive client (1Hz reaction) and a reasonable cpu load (100Hz data rate)
        self._mock_update_rate = min(self._mock_update_rate, self.MAX_MOCK_UPDATE_RATE_HZ)
        self._mock_update_rate = max(self._mock_update_rate, self.MIN_MOCK_UPDATE_RATE_HZ)

        if self.session_config.extended:
            return self._metadata
        else:
            return unextend(self._metadata)

    def start_session(self, recorder: Optional[Recorder] = None) -> None:
        self._assert_session_setup()

        if self.session_is_started:
            raise ClientError("Session is already started.")

        self._recorder_start(recorder)
        self._create_rate_stats_calc()
        self._session_is_started = True
        self._start_time = time.monotonic()
        self._mock_next_data_time = self._start_time

    def get_next(self) -> Union[Result, list[dict[int, Result]]]:
        self._assert_session_started()

        if self._metadata is None:
            raise RuntimeError(f"{self} has no metadata")

        if self._session_config is None:
            raise RuntimeError(f"{self} has no session config")

        extended_results = self._session_config_to_result(self.session_config)

        delta = self._mock_next_data_time - time.perf_counter()
        if delta > 0:
            time.sleep(delta)

        self._mock_next_data_time += 1 / self._mock_update_rate

        self._recorder_sample(extended_results)
        self._update_rate_stats_calc(extended_results)
        return self._return_results(extended_results)

    def stop_session(self) -> Any:
        self._assert_session_started()
        self._session_is_started = False
        self._rate_stats_calc = None

        return self._recorder_stop()

    def disconnect(self) -> None:
        if not self._connected:
            raise ClientError("Client is already disconnected")

        if self.session_is_started:
            _ = self.stop_session()

        self._metadata = None
        self._connected = False

    @property
    def connected(self) -> bool:
        return self._connected

    @property
    def server_info(self) -> ServerInfo:
        self._assert_connected()
        return self.MOCK_SERVER_INFO



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from . import plugin_components, utils
from .app_model_viewer import AppModelViewer
from .main_window import MainWindow
from .misc import BUTTON_ICON_COLOR
from .plugin_components import (
    AttrsConfigEditor,
    GridGroupBox,
    HorizontalGroupBox,
    MiscErrorView,
    PidgetFactoryMapping,
    SessionConfigEditor,
    SmartMetadataView,
    SmartPerfCalcView,
    TwoSensorIdsEditor,
    VerticalGroupBox,
    pidgets,
)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import typing as t

import attrs
import numpy as np

from acconeer.exptool import a121


@attrs.frozen(kw_only=True)
class _RateStats:
    rate: float = attrs.field(default=np.nan)
    rate_warning: bool = attrs.field(default=False)
    jitter: float = attrs.field(default=np.nan)
    jitter_warning: bool = attrs.field(default=False)


class _RateCalculator:
    _JITTER_WARNING_LIMIT = 1.0e-3  # Based on testing

    stats: _RateStats

    def __init__(
        self,
        session_config: a121.SessionConfig,
        metadata: t.Union[a121.Metadata, list[dict[int, a121.Metadata]]],
    ) -> None:
        if isinstance(metadata, list):
            metadata = next(a121.iterate_extended_structure_values(metadata))

        self.metadata = metadata

        self.session_config = session_config  # TODO: For performance calculator based warning
        self.sensor_config = next(a121.iterate_extended_structure_values(session_config.groups))

        self.last_result: t.Optional[a121.Result] = None
        self.time_fifo = np.full(200, np.nan)
        self.tick_fifo = np.full(200, np.nan)
        self.stats = _RateStats()

    def update(
        self,
        result: t.Union[a121.Result, list[dict[int, a121.Result]]],
    ) -> None:
        if isinstance(result, list):
            result = next(a121.iterate_extended_structure_values(result))

        last_result = self.last_result
        self.last_result = result

        if last_result is None:
            return

        delta_time = result.tick_time - last_result.tick_time
        self.time_fifo = np.roll(self.time_fifo, -1)
        self.time_fifo[-1] = delta_time

        delta_tick = result.tick - last_result.tick
        self.tick_fifo = np.roll(self.tick_fifo, -1)
        self.tick_fifo[-1] = delta_tick

        rate = 1.0 / np.nanmean(self.time_fifo)
        jitter = np.nanstd(self.time_fifo)

        rate_warning = result.frame_delayed

        if self.session_config.update_rate is not None:
            limit = self.metadata.tick_period * 1.01 + 10
            rate_warning |= np.nanmean(self.tick_fifo) > limit

        jitter_warning = jitter > self._JITTER_WARNING_LIMIT

        self.stats = _RateStats(
            rate=float(rate),
            rate_warning=bool(rate_warning),
            jitter=float(jitter),
            jitter_warning=bool(jitter_warning),
        )



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import logging
from typing import List

import requests

from bs4 import BeautifulSoup


DEV_LICENSE_URL = "https://developer.acconeer.com/software-license-agreement/"

DEV_LICENSE_DEFAULT_HEADER = "SOFTWARE LICENSE AGREEMENT"
DEV_LICENSE_DEFAULT_SUBHEADER = "LIMITED LICENSE AGREEMENT FOR Acconeer MATERIALS"
DEV_LICENSE_DEFAULT_CONFIRMATION_MSG = (
    "Please read carefully the terms and conditions at "
    "https://developer.acconeer.com/software-license-agreement/"
)

log = logging.getLogger(__name__)


class DevLicense:
    """Acconeer developer license abstration"""

    def __init__(self, license_url: str = DEV_LICENSE_URL) -> None:
        self.license_url = license_url
        self.html = None

    def load(self) -> None:
        """Loads the license HTML document"""
        try:
            self.html = BeautifulSoup(requests.get(self.license_url).text, "html.parser")
        except Exception as e:
            log.warn(str(e))
            self.html = None

    def get_header(self) -> str:
        """Get the HTML header as a string

        :returns: HTML header.

        """
        header = DEV_LICENSE_DEFAULT_HEADER

        if self.html is not None:
            header_element = self.html.find("h1")
            if header_element is not None:
                header = header_element.contents[0].string

        return header

    def get_header_element(self) -> str:
        """Get the HTML header element as a string

        :returns: HTML header element.

        """
        header_elem = "<h1>" + DEV_LICENSE_DEFAULT_HEADER + "</h1>"

        if self.html is not None:
            header_elem = str(self.html.find("h1"))

        return header_elem

    def get_subheader(self) -> str:
        """Get the HTML subheader as a string

        :returns: HTML subheader.

        """
        subheader = DEV_LICENSE_DEFAULT_SUBHEADER

        if self.html is not None:
            subheader_element = self.html.find("h3")
            if subheader_element is not None:
                subheader = subheader_element.contents[0].string

        return subheader

    def get_subheader_element(self) -> str:
        """Get the HTML subheader element as a string

        :returns: HTML subheader element.

        """
        subheader_elem = "<h3>" + DEV_LICENSE_DEFAULT_SUBHEADER + "</h3>"

        if self.html is not None:
            subheader_elem = str(self.html.find("h3"))

        return subheader_elem

    def get_content(self) -> List[str]:
        """Get the HTML paragraph content as a list of strings

        :returns: HTML paragraphs.

        """
        paragraphs = []

        if self.html is not None:
            for p in self.html.select("p"):
                if p is not None and p.string is not None:
                    # The "\xa0" is a non-breaking space (&nbsp;) that is not
                    # escaped by BeautifulSoup and hence filtered out here.
                    paragraphs.append(p.string.replace("\xa0", " "))
        else:
            paragraphs.append(DEV_LICENSE_DEFAULT_CONFIRMATION_MSG)

        return paragraphs

    def get_content_elements(self) -> List[str]:
        """Get the HTML paragraph elements as a list of strings

        :returns: HTML paragraphs elements.

        """
        paragraphs = []

        if self.html is not None:
            paragraphs = [str(p) for p in self.html.select("p")]
        else:
            paragraphs.append("<p></p>")
            paragraphs.append(DEV_LICENSE_DEFAULT_CONFIRMATION_MSG)

        return paragraphs



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from acconeer.exptool.a111.algo import ModuleFamily, ModuleInfo

from ._processor import ProcessingConfiguration, Processor, get_sensor_config
from .calibration import (
    EnvelopeCalibration,
    EnvelopeCalibrationConfiguration,
    EnvelopeCalibrationMapper,
)
from .ui import PGUpdater


module_info = ModuleInfo(
    key="tank_level_short",
    label="Tank level short range (envelope)",
    pg_updater=PGUpdater,
    processing_config_class=ProcessingConfiguration,
    module_family=ModuleFamily.EXAMPLE,
    sensor_config_class=get_sensor_config,
    processor=Processor,
    multi_sensor=False,
    docs_url=None,
    calibration_class=EnvelopeCalibration,
    calibration_config_class=EnvelopeCalibrationConfiguration,
    calibration_mapper=EnvelopeCalibrationMapper,
)



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved
from __future__ import annotations

import typing as t

import h5py
import numpy as np

from ._processors import ProcessorResult


S = t.TypeVar("S")
T = t.TypeVar("T")

_ALL_RESULT_FIELDS: t.Final = (
    "intra_presence_score",
    "inter_presence_score",
    "presence_distance",
    "presence_detected",
    "intra",
    "inter",
    # "extra_result" ignored by default
)


class PhonyNoneSeries:
    def __next__(self) -> None:
        return None

    def __iter__(self) -> PhonyNoneSeries:
        return self


class ProcessorResultListH5Serializer:
    """
    Reads or writes a presence ProcessorResult from/to a given h5py.Group
    """

    def __init__(
        self,
        group: h5py.Group,
        fields: t.Sequence[str] = _ALL_RESULT_FIELDS,
        allow_missing_fields: bool = False,
    ) -> None:
        """
        :param destination: The H5 group that will have datasets read/written to it
        :param fields: The fields to serialize. Has no effect on de-serialization
        :param allow_missing_fields:
            Whether it's acceptable to break type safety during de-serialization
            by replacing missing fields by ``None``. Has no effect on serialization.
        """
        self.group = group
        self.fields = fields
        self.allow_missing_fields = allow_missing_fields

    def serialize(self, results: t.List[ProcessorResult]) -> None:
        if "extra_result" in self.fields:
            raise NotImplementedError(
                "'extra_results' are not serializable at the moment."
                + "Skip it by specifying which fields to serialize"
            )

        if "intra_presence_score" in self.fields:
            self.group.create_dataset(
                "intra_presence_score",
                dtype=float,
                data=np.array([res.intra_presence_score for res in results]),
                track_times=False,
            )

        if "inter_presence_score" in self.fields:
            self.group.create_dataset(
                "inter_presence_score",
                dtype=float,
                data=np.array([res.inter_presence_score for res in results]),
                track_times=False,
            )

        if "presence_distance" in self.fields:
            self.group.create_dataset(
                "presence_distance",
                dtype=float,
                data=np.array([res.presence_distance for res in results]),
                track_times=False,
            )

        if "presence_detected" in self.fields:
            self.group.create_dataset(
                "presence_detected",
                dtype=bool,
                data=np.array([res.presence_detected for res in results]),
                track_times=False,
            )

        if "intra" in self.fields:
            self.group.create_dataset(
                "intra",
                dtype=float,
                data=np.array([res.intra for res in results]),
                track_times=False,
            )

        if "inter" in self.fields:
            self.group.create_dataset(
                "inter",
                dtype=float,
                data=np.array([res.inter for res in results]),
                track_times=False,
            )

    def deserialize(self, _: None) -> t.List[ProcessorResult]:
        if "extra_result" in self.fields:
            raise NotImplementedError(
                "'extra_results' are not serializable at the moment."
                + "Skip it by specifying which fields to serialize"
            )

        groups = (
            self.group.get("intra_presence_score", PhonyNoneSeries()),
            self.group.get("inter_presence_score", PhonyNoneSeries()),
            self.group.get("presence_distance", PhonyNoneSeries()),
            self.group.get("presence_detected", PhonyNoneSeries()),
            self.group.get("intra", PhonyNoneSeries()),
            self.group.get("inter", PhonyNoneSeries()),
        )

        if all(isinstance(group, PhonyNoneSeries) for group in groups):
            return []

        return [
            ProcessorResult(
                intra_presence_score=intra_presence_score,
                inter_presence_score=inter_presence_score,
                presence_distance=distance,
                presence_detected=detected,
                intra=intra,
                inter=inter,
                extra_result=None,  # type: ignore[arg-type]
            )
            for (
                intra_presence_score,
                inter_presence_score,
                distance,
                detected,
                intra,
                inter,
            ) in zip(*groups)
        ]



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

import copy
from typing import Any, Tuple

import numpy as np
import numpy.typing as npt
from scipy.signal import butter

from acconeer.exptool import a121


ENVELOPE_FWHM_M = {
    a121.Profile.PROFILE_1: 0.04,
    a121.Profile.PROFILE_2: 0.07,
    a121.Profile.PROFILE_3: 0.14,
    a121.Profile.PROFILE_4: 0.19,
    a121.Profile.PROFILE_5: 0.32,
}
APPROX_BASE_STEP_LENGTH_M = 2.5e-3
# Parameter of signal temperature model.
SIGNAL_TEMPERATURE_MODEL_PARAMETER = {
    a121.Profile.PROFILE_1: 67.0,
    a121.Profile.PROFILE_2: 85.0,
    a121.Profile.PROFILE_3: 86.0,
    a121.Profile.PROFILE_4: 99.0,
    a121.Profile.PROFILE_5: 104.0,
}
# Largest measurable distance per PRF.
MAX_MEASURABLE_DIST_M = {
    a121.PRF.PRF_19_5_MHz: 3.1,
    a121.PRF.PRF_15_6_MHz: 5.1,
    a121.PRF.PRF_13_0_MHz: 7.0,
    a121.PRF.PRF_8_7_MHz: 12.7,
    a121.PRF.PRF_6_5_MHz: 18.5,
    a121.PRF.PRF_5_2_MHz: 28.8,
}
# Slope and interception of linear noise temperature model.
NOISE_TEMPERATURE_MODEL_PARAMETER = [-0.00275, 0.98536]

SPEED_OF_LIGHT = 299792458
RADIO_FREQUENCY = 60.5e9
WAVELENGTH = SPEED_OF_LIGHT / RADIO_FREQUENCY
PERCEIVED_WAVELENGTH = WAVELENGTH / 2


def get_distances_m(
    config: a121.SensorConfig,
    metadata: a121.Metadata,
) -> Tuple[npt.NDArray[np.float_], float]:
    points = np.arange(config.num_points) * config.step_length + config.start_point
    distances_m = points * metadata.base_step_length_m
    step_length_m = config.step_length * metadata.base_step_length_m
    return distances_m, step_length_m


def get_approx_sweep_rate(config: a121.SensorConfig) -> float:
    ppp = {1: 24, 2: 20, 3: 16, 4: 16, 5: 16}[config.profile.value]

    n = 3 * ppp + config.num_points * config.hwaas * ppp

    return config.prf.frequency / n


def get_approx_fft_vels(config: a121.SensorConfig) -> Tuple[npt.NDArray, float]:
    sweep_rate = get_approx_sweep_rate(config)
    if config.sweep_rate is not None:
        sweep_rate = min([sweep_rate, config.sweep_rate])

    spf = config.sweeps_per_frame
    f_res = 1 / spf
    freqs = np.fft.fftshift(np.fft.fftfreq(spf))  # type: ignore[call-overload]
    f_to_v = 2.5e-3 * sweep_rate
    return freqs * f_to_v, f_res * f_to_v


def interpolate_peaks(
    abs_sweep: npt.NDArray[np.float_],
    peak_idxs: list[int],
    start_point: int,
    step_length: int,
    step_length_m: float,
) -> Tuple[list[float], list[float]]:
    """Quadratic interpolation around a peak using the amplitudes of the peak and its two
    neghbouring points.

    Derivation:
    https://math.stackexchange.com/questions/680646/get-polynomial-function-from-3-points

    :param abs_sweep: Absolute value of mean sweep.
    :param peak_idxs: List containing indexes of identified peaks.
    :param start_point: Start point.
    :param step_length: Step length in points.
    :param step_length_m: Step length in meters.
    """
    estimated_distances = []
    estimated_amplitudes = []
    for peak_idx in peak_idxs:
        x = np.arange(peak_idx - 1, peak_idx + 2, 1)
        y = abs_sweep[peak_idx - 1 : peak_idx + 2]
        a = (x[0] * (y[2] - y[1]) + x[1] * (y[0] - y[2]) + x[2] * (y[1] - y[0])) / (
            (x[0] - x[1]) * (x[0] - x[2]) * (x[1] - x[2])
        )
        b = (y[1] - y[0]) / (x[1] - x[0]) - a * (x[0] + x[1])
        c = y[0] - a * x[0] ** 2 - b * x[0]
        peak_loc = -b / (2 * a)
        estimated_distances.append((start_point + peak_loc * step_length) * step_length_m)
        estimated_amplitudes.append(a * peak_loc**2 + b * peak_loc + c)
    return estimated_distances, estimated_amplitudes


def find_peaks(abs_sweep: npt.NDArray[np.float_], threshold: npt.NDArray[np.float_]) -> list[int]:
    """Identifies peaks above threshold.

    A peak is defined as a point with greater value than its two neighbouring points and all
    three points are above the threshold.

    :param abs_sweep: Absolute value of mean sweep.
    :param threshold: Array of values, defining the threshold throughout the sweep.
    """
    if threshold is None:
        raise ValueError
    found_peaks = []
    d = 1
    N = len(abs_sweep)
    while d < (N - 1):
        if np.isnan(threshold[d - 1]):
            d += 1
            continue
        if np.isnan(threshold[d + 1]):
            break
        if abs_sweep[d] <= threshold[d]:
            d += 2
            continue
        if abs_sweep[d - 1] <= threshold[d - 1]:
            d += 1
            continue
        if abs_sweep[d - 1] >= abs_sweep[d]:
            d += 1
            continue
        d_upper = d + 1
        while True:
            if (d_upper) >= (N - 1):
                break
            if np.isnan(threshold[d_upper]):
                break
            if abs_sweep[d_upper] <= threshold[d_upper]:
                break
            if abs_sweep[d_upper] > abs_sweep[d]:
                break
            elif abs_sweep[d_upper] < abs_sweep[d]:
                found_peaks.append(int(np.argmax(abs_sweep[d:d_upper]) + d))
                break
            else:
                d_upper += 1
        d = d_upper
    return found_peaks


def get_temperature_adjustment_factors(
    temperature_diff: int, profile: a121.Profile
) -> Tuple[float, float]:
    """Calculate temperature compensation for mean sweep and background noise(tx off) standard
    deviation.

    The signal adjustment model is follows 2 ** (temperature_diff / model_parameter), where
    model_parameter reflects the temperature difference relative the reference temperature,
    required for the amplitude to double/halve.

    The noise adjustment is a linear function of the temperature difference, calibrated using
    noise-normalized data, generalizing to different sensor configurations.
    """
    signal_adjustment_factor = 2 ** (
        temperature_diff / SIGNAL_TEMPERATURE_MODEL_PARAMETER[profile]
    )
    noise_adjustment_factor = (
        NOISE_TEMPERATURE_MODEL_PARAMETER[0] * temperature_diff
        + NOISE_TEMPERATURE_MODEL_PARAMETER[1]
    )
    return (signal_adjustment_factor, noise_adjustment_factor)


def get_distance_filter_coeffs(profile: a121.Profile, step_length: int) -> Any:
    """Calculates the iir coefficients corresponding to a matched filter, based on the profile and
    the step length.
    """
    wnc = APPROX_BASE_STEP_LENGTH_M * step_length / (ENVELOPE_FWHM_M[profile])
    return butter(N=2, Wn=wnc)


def get_distance_filter_edge_margin(profile: a121.Profile, step_length: int) -> int:
    """Calculates the number of points required for filter initialization when performing
    distance filtering, using the filter coefficients supplied by the function
    get_distance_filter_coeffs.
    """
    return int(_safe_ceil(ENVELOPE_FWHM_M[profile] / (APPROX_BASE_STEP_LENGTH_M * step_length)))


def select_prf(breakpoint: int, profile: a121.Profile) -> a121.PRF:
    """Calculates the highest possible PRF for the given breakpoint.

    :param breakpoint: Distance in the unit of points.
    :param profile: Profile.
    """
    max_meas_dist_m = copy.copy(MAX_MEASURABLE_DIST_M)

    if a121.PRF.PRF_19_5_MHz in max_meas_dist_m and profile != a121.Profile.PROFILE_1:
        del max_meas_dist_m[a121.PRF.PRF_19_5_MHz]

    breakpoint_m = breakpoint * APPROX_BASE_STEP_LENGTH_M
    viable_prfs = [prf for prf, max_dist_m in max_meas_dist_m.items() if breakpoint_m < max_dist_m]
    return sorted(viable_prfs, key=lambda prf: prf.frequency)[-1]


def _safe_ceil(x: float) -> float:
    """Perform safe ceil.

    Implementation of ceil function, compatible with float representation in C.
    """
    return float(f"{x:.16g}")



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from typing import Any, Callable, Generic, Optional, Type

import attrs

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool._bs_thread import BSThread, BSThreadDiedException  # type: ignore[import]
from acconeer.exptool.a121 import algo
from acconeer.exptool.a121.algo._base import InputT, MetadataT, ProcessorConfigT, ResultT
from acconeer.exptool.app.new import GeneralMessage

from ._null_app_model import NullAppModel
from .processor import GenericProcessorPlotPluginBase


def processor_main(
    *,
    processor_cls: Type[algo.GenericProcessorBase[InputT, ProcessorConfigT, ResultT, MetadataT]],
    processor_config_cls: Type[ProcessorConfigT],
    plot_plugin: Type[GenericProcessorPlotPluginBase[ResultT, MetadataT]],
    sensor_config_getter: Callable[[], a121.SensorConfig],
    _blinkstick_updater_cls: Optional[Any] = None,
) -> None:
    parser = a121.ExampleArgumentParser()
    parser.add_argument("--sensor", type=int, default=1)
    args = parser.parse_args()
    et.utils.config_logging(args)

    client = a121.Client(**a121.get_client_args(args))
    client.connect()

    sensor_config = sensor_config_getter()
    session_config = a121.SessionConfig({args.sensor: sensor_config})

    metadata = client.setup_session(session_config)

    processor_config = processor_config_cls()

    processor = processor_cls(
        sensor_config=sensor_config,
        metadata=metadata,  # type: ignore[arg-type]
        processor_config=processor_config,
    )

    pg_updater = ProcessorPGUpdater[ResultT, MetadataT](
        plot_plugin=plot_plugin,
        sensor_config=sensor_config,
        metadata=metadata,  # type: ignore[arg-type]
    )
    pg_process = et.PGProcess(pg_updater)  # type: ignore[attr-defined]
    pg_process.start()

    if _blinkstick_updater_cls is None:
        bs_process = None
    else:
        bs_updater = _blinkstick_updater_cls()
        bs_process = BSThread(bs_updater)
        bs_process.start()

    client.start_session()

    interrupt_handler = et.utils.ExampleInterruptHandler()
    print("Press Ctrl-C to end session")

    while not interrupt_handler.got_signal:
        result = client.get_next()

        processor_result = processor.process(result)  # type: ignore[arg-type]

        try:
            pg_process.put_data(processor_result)

            if bs_process is not None:
                bs_process.put_data(processor_result)
        except (et.PGProccessDiedException, BSThreadDiedException):  # type: ignore[attr-defined]
            break

    print("Disconnecting...")

    pg_process.close()

    if bs_process is not None:
        bs_process.close()

    client.disconnect()


@attrs.mutable(kw_only=True, slots=False)
class ProcessorPGUpdater(Generic[ResultT, MetadataT]):
    plot_plugin: Type[GenericProcessorPlotPluginBase[ResultT, MetadataT]] = attrs.field()
    sensor_config: a121.SensorConfig = attrs.field()
    metadata: MetadataT = attrs.field()
    plot_plugin_obj: Optional[GenericProcessorPlotPluginBase[ResultT, MetadataT]] = attrs.field(
        default=None, init=False
    )

    def setup(self, win: pg.GraphicsLayout) -> None:
        self.plot_plugin_obj = self.plot_plugin(plot_layout=win, app_model=NullAppModel())
        self.plot_plugin_obj.handle_message(
            GeneralMessage(
                name="setup", kwargs=dict(sensor_config=self.sensor_config, metadata=self.metadata)
            )
        )

    def update(self, data: Any) -> None:
        if self.plot_plugin_obj is None:
            raise RuntimeError

        self.plot_plugin_obj.handle_message(GeneralMessage(name="plot", data=data))
        self.plot_plugin_obj.draw()



# Copyright (c) Acconeer AB, 2022
# All rights reserved
from __future__ import annotations

from typing import Optional

from PySide6 import QtCore
from PySide6.QtWidgets import QFrame, QScrollArea, QVBoxLayout, QWidget


class ScrollAreaDecorator(QScrollArea):
    def __init__(self, decoratee: QWidget) -> None:
        """Puts `decoratee` in a frameless QScrollArea with no horizontal scroll."""
        super().__init__()

        self.setWidgetResizable(True)
        self.setFrameShape(QFrame.NoFrame)
        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.horizontalScrollBar().setEnabled(False)

        self.setWidget(decoratee)


class TopAlignDecorator(QWidget):
    def __init__(self, decoratee: QWidget) -> None:
        """Top-aligns `decoratee` by using QVBoxLayout and its stretch-functionality"""
        super().__init__()

        layout = QVBoxLayout(self)
        self.setLayout(layout)

        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(decoratee)
        layout.addStretch(1)


class HorizontalSeparator(QFrame):
    def __init__(self, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)
        self.setFrameShape(QFrame.HLine)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import abc
import logging
from typing import Generic, Optional, Type

import qtawesome as qta

from PySide6.QtWidgets import QPushButton, QVBoxLayout, QWidget

from acconeer.exptool import a121
from acconeer.exptool.a121.algo._base import ProcessorConfigT
from acconeer.exptool.a121.algo._plugins._a121 import A121ViewPluginBase
from acconeer.exptool.app.new import (
    BUTTON_ICON_COLOR,
    AppModel,
    AttrsConfigEditor,
    GeneralMessage,
    GridGroupBox,
    MiscErrorView,
    PidgetFactoryMapping,
    PluginState,
    SessionConfigEditor,
    SmartMetadataView,
    SmartPerfCalcView,
)

from .backend_plugin import ProcessorBackendPluginSharedState


log = logging.getLogger(__name__)


class ProcessorViewPluginBase(A121ViewPluginBase, Generic[ProcessorConfigT]):
    def __init__(self, app_model: AppModel, view_widget: QWidget) -> None:
        super().__init__(app_model=app_model, view_widget=view_widget)

        sticky_layout = QVBoxLayout()
        sticky_layout.setContentsMargins(0, 0, 0, 0)
        scrolly_layout = QVBoxLayout()
        scrolly_layout.setContentsMargins(0, 0, 0, 0)

        self.start_button = QPushButton(
            qta.icon("fa5s.play-circle", color=BUTTON_ICON_COLOR),
            "Start measurement",
            self.sticky_widget,
        )
        self.stop_button = QPushButton(
            qta.icon("fa5s.stop-circle", color=BUTTON_ICON_COLOR),
            "Stop",
            self.sticky_widget,
        )
        self.start_button.clicked.connect(self._send_start_request)
        self.stop_button.clicked.connect(self._send_stop_request)

        self.start_button.setShortcut("space")
        self.start_button.setToolTip("Starts the session.\n\nShortcut: Space")
        self.stop_button.setShortcut("space")
        self.stop_button.setToolTip("Stops the session.\n\nShortcut: Space")

        button_group = GridGroupBox("Controls", parent=self.sticky_widget)
        button_group.layout().addWidget(self.start_button, 0, 0)
        button_group.layout().addWidget(self.stop_button, 0, 1)
        sticky_layout.addWidget(button_group)

        self.metadata_view = SmartMetadataView(parent=self.scrolly_widget)
        scrolly_layout.addWidget(self.metadata_view)

        self.perf_calc_view = SmartPerfCalcView(parent=self.scrolly_widget)
        scrolly_layout.addWidget(self.perf_calc_view)

        self.processor_config_editor = AttrsConfigEditor[ProcessorConfigT](
            title="Processor parameters",
            factory_mapping=self.get_pidget_mapping(),
            parent=self.scrolly_widget,
        )
        self.processor_config_editor.sig_update.connect(self._on_processor_config_update)
        scrolly_layout.addWidget(self.processor_config_editor)

        self.misc_error_view = MiscErrorView(self.scrolly_widget)
        scrolly_layout.addWidget(self.misc_error_view)

        self.session_config_editor = SessionConfigEditor(
            self.supports_multiple_subsweeps(), self.scrolly_widget
        )
        self.session_config_editor.sig_update.connect(self._on_session_config_update)
        scrolly_layout.addWidget(self.session_config_editor)

        self.sticky_widget.setLayout(sticky_layout)
        self.scrolly_widget.setLayout(scrolly_layout)

    def _on_session_config_update(self, session_config: a121.SessionConfig) -> None:
        self.app_model.put_backend_plugin_task(
            "update_session_config", {"session_config": session_config}
        )

    def _on_processor_config_update(self, processor_config: ProcessorConfigT) -> None:
        self.app_model.put_backend_plugin_task(
            "update_processor_config", {"processor_config": processor_config}
        )

    def handle_message(self, message: GeneralMessage) -> None:
        if message.name == "sync":
            log.debug(f"{type(self).__name__} syncing")

            self.session_config_editor.sync()
            self.processor_config_editor.sync()
        else:
            raise RuntimeError("Unknown message")

    def on_backend_state_update(
        self, backend_plugin_state: Optional[ProcessorBackendPluginSharedState]
    ) -> None:
        if backend_plugin_state is None:
            self.session_config_editor.set_data(None)
            self.processor_config_editor.set_data(None)
            self.metadata_view.update(None)
            self.perf_calc_view.update()
        else:
            assert isinstance(backend_plugin_state, ProcessorBackendPluginSharedState)
            assert isinstance(
                backend_plugin_state.processor_config, self.get_processor_config_cls()
            )

            self.session_config_editor.set_data(backend_plugin_state.session_config)
            self.processor_config_editor.set_data(backend_plugin_state.processor_config)
            self.metadata_view.update(backend_plugin_state.metadata)
            self.perf_calc_view.update(
                backend_plugin_state.session_config,
                backend_plugin_state.metadata,
            )

            if backend_plugin_state.processor_config is not None:
                results = (
                    backend_plugin_state.processor_config._collect_validation_results(
                        backend_plugin_state.session_config
                    )
                    + backend_plugin_state.session_config._collect_validation_results()
                )

                not_handled = self.session_config_editor.handle_validation_results(results)

                not_handled = self.processor_config_editor.handle_validation_results(not_handled)

                not_handled = self.misc_error_view.handle_validation_results(not_handled)

                assert not_handled == []

    def on_app_model_update(self, app_model: AppModel) -> None:
        self.session_config_editor.setEnabled(app_model.plugin_state == PluginState.LOADED_IDLE)
        self.processor_config_editor.setEnabled(app_model.plugin_state == PluginState.LOADED_IDLE)

        self.start_button.setEnabled(
            app_model.is_ready_for_session() and app_model.backend_plugin_state.ready
        )

        self.stop_button.setEnabled(app_model.plugin_state == PluginState.LOADED_BUSY)

        if app_model.backend_plugin_state:
            self.session_config_editor.set_selected_sensor(
                app_model.backend_plugin_state.session_config.sensor_id,
                self.app_model.connected_sensors,
            )

    @classmethod
    def supports_multiple_subsweeps(self) -> bool:
        return False

    @classmethod
    @abc.abstractmethod
    def get_pidget_mapping(cls) -> PidgetFactoryMapping:
        pass

    @classmethod
    @abc.abstractmethod
    def get_processor_config_cls(cls) -> Type[ProcessorConfigT]:
        pass



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from .entities import (
    INT_16_COMPLEX,
    PRF,
    ClientInfo,
    Criticality,
    IdleState,
    Metadata,
    PersistentRecord,
    Profile,
    Record,
    Result,
    SensorCalibration,
    SensorConfig,
    SensorInfo,
    ServerInfo,
    SessionConfig,
    StackedResults,
    SubsweepConfig,
    ValidationError,
    ValidationResult,
    ValidationWarning,
    complex_array_to_int16_complex,
    int16_complex_array_to_complex,
)
from .mediators import ClientBase, ClientError, Recorder
from .peripherals import (
    _H5PY_STR_DTYPE,
    Client,
    H5Record,
    H5Recorder,
    InMemoryRecord,
    RecordError,
    ServerError,
    load_record,
    open_record,
    save_record,
    save_record_to_h5,
)
from .utils import (
    iterate_extended_structure,
    iterate_extended_structure_values,
    zip3_extended_structures,
    zip_extended_structures,
)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from ._detector import DetectorBackendPluginBase, DetectorPlotPluginBase, DetectorViewPluginBase
from ._processor_main import processor_main
from .processor import (
    ProcessorBackendPluginBase,
    ProcessorPlotPluginBase,
    ProcessorPluginPreset,
    ProcessorPluginSpec,
    ProcessorViewPluginBase,
)



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from __future__ import annotations

import copy
import enum
import warnings
from typing import Any, Dict, List, Optional, Tuple

import attrs
import h5py
import numpy as np
import numpy.typing as npt

from acconeer.exptool import a121
from acconeer.exptool.a121._h5_utils import _create_h5_string_dataset
from acconeer.exptool.a121.algo import (
    APPROX_BASE_STEP_LENGTH_M,
    ENVELOPE_FWHM_M,
    AlgoBase,
    AlgoConfigBase,
    get_distance_filter_edge_margin,
    select_prf,
)

from ._aggregator import Aggregator, AggregatorConfig, PeakSortingMethod, ProcessorSpec
from ._processors import (
    DEFAULT_FIXED_THRESHOLD_VALUE,
    DEFAULT_THRESHOLD_SENSITIVITY,
    MeasurementType,
    Processor,
    ProcessorConfig,
    ProcessorContext,
    ProcessorMode,
    ProcessorResult,
    ThresholdMethod,
    calculate_bg_noise_std,
    calculate_loopback_peak_location,
)


@attrs.frozen(kw_only=True)
class DetectorStatus:
    detector_state: DetailedStatus
    ready_to_start: bool


class DetailedStatus(enum.Enum):
    OK = enum.auto()
    END_LESSER_THAN_START = enum.auto()
    SENSOR_IDS_NOT_UNIQUE = enum.auto()
    CONTEXT_MISSING = enum.auto()
    CALIBRATION_MISSING = enum.auto()
    CONFIG_MISMATCH = enum.auto()
    INVALID_DETECTOR_CONFIG_RANGE = enum.auto()


@attrs.frozen(kw_only=True)
class SubsweepGroupPlan:
    step_length: int = attrs.field()
    breakpoints: list[int] = attrs.field()
    profile: a121.Profile = attrs.field()
    hwaas: list[int] = attrs.field()


Plan = Dict[MeasurementType, List[SubsweepGroupPlan]]


@attrs.mutable(kw_only=True)
class DetectorContext(AlgoBase):
    single_sensor_contexts: Dict[int, SingleSensorContext] = attrs.field(default=None)
    _GROUP_NAME = "sensor_id_"

    @property
    def sensor_ids(self) -> Optional[list[int]]:
        if self.single_sensor_contexts:
            return list(self.single_sensor_contexts.keys())
        else:
            return None

    def to_h5(self, group: h5py.Group) -> None:
        if self.single_sensor_contexts is not None:
            for sensor_id, context in self.single_sensor_contexts.items():
                context.to_h5(group.create_group(self._GROUP_NAME + str(sensor_id)))

    @classmethod
    def from_h5(cls, group: h5py.Group) -> DetectorContext:
        context_dict = {}

        for key in group.keys():
            if cls._GROUP_NAME in key:
                sensor_id = int(key.split("_")[-1])
                context_dict[sensor_id] = SingleSensorContext.from_h5(group[key])

        return DetectorContext(single_sensor_contexts=context_dict)


@attrs.mutable(kw_only=True)
class SingleSensorContext(AlgoBase):
    loopback_peak_location_m: Optional[float] = attrs.field(default=None)
    direct_leakage: Optional[npt.NDArray[np.complex_]] = attrs.field(default=None)
    phase_jitter_comp_reference: Optional[npt.NDArray[np.float_]] = attrs.field(default=None)
    recorded_thresholds_mean_sweep: Optional[List[npt.NDArray[np.float_]]] = attrs.field(
        default=None
    )
    recorded_thresholds_noise_std: Optional[List[List[np.float_]]] = attrs.field(default=None)
    bg_noise_std: Optional[List[List[float]]] = attrs.field(default=None)
    session_config_used_during_calibration: Optional[a121.SessionConfig] = attrs.field(
        default=None
    )
    reference_temperature: Optional[int] = attrs.field(default=None)
    sensor_calibration: Optional[a121.SensorCalibration] = attrs.field(default=None)
    # TODO: Make recorded_thresholds Optional[List[Optional[npt.NDArray[np.float_]]]]

    def to_h5(self, group: h5py.Group) -> None:
        for k, v in attrs.asdict(self, recurse=False).items():
            if k in [
                "recorded_thresholds_mean_sweep",
                "recorded_thresholds_noise_std",
                "bg_noise_std",
            ]:
                continue

            if v is None:
                continue

            if isinstance(v, a121.SessionConfig):
                _create_h5_string_dataset(group, k, v.to_json())
            elif isinstance(v, a121.SensorCalibration):
                sensor_calibration_group = group.create_group("sensor_calibration")
                v.to_h5(sensor_calibration_group)
            elif isinstance(v, np.ndarray) or isinstance(v, float) or isinstance(v, int):
                group.create_dataset(k, data=v, track_times=False)
            else:
                raise RuntimeError(
                    f"Unexpected {type(self).__name__} field '{k}' of type '{type(v)}'"
                )

        if self.recorded_thresholds_mean_sweep is not None:
            recorded_thresholds_mean_sweep_group = group.create_group(
                "recorded_thresholds_mean_sweep"
            )

            for i, v in enumerate(self.recorded_thresholds_mean_sweep):
                recorded_thresholds_mean_sweep_group.create_dataset(
                    f"index_{i}", data=v, track_times=False
                )

        if self.recorded_thresholds_noise_std is not None:
            recorded_thresholds_std_group = group.create_group("recorded_thresholds_noise_std")

            for i, v in enumerate(self.recorded_thresholds_noise_std):
                recorded_thresholds_std_group.create_dataset(
                    f"index_{i}", data=v, track_times=False
                )

        if self.bg_noise_std is not None:
            bg_noise_std_group = group.create_group("bg_noise_std")

            for i, v in enumerate(self.bg_noise_std):
                bg_noise_std_group.create_dataset(f"index_{i}", data=v, track_times=False)

    @classmethod
    def from_h5(cls, group: h5py.Group) -> SingleSensorContext:
        context_dict = {}

        unknown_keys = set(group.keys()) - set(attrs.fields_dict(SingleSensorContext).keys())
        if unknown_keys:
            raise Exception(f"Unknown field(s) in stored context: {unknown_keys}")

        field_map = {
            "loopback_peak_location_m": None,
            "direct_leakage": None,
            "reference_temperature": None,
            "phase_jitter_comp_reference": None,
            "session_config_used_during_calibration": a121.SessionConfig.from_json,
        }
        for k, func in field_map.items():
            try:
                v = group[k][()]
            except KeyError:
                continue

            context_dict[k] = func(v) if func else v

        if "recorded_thresholds_mean_sweep" in group:
            mean_sweeps = _get_group_items(group["recorded_thresholds_mean_sweep"])
            context_dict["recorded_thresholds_mean_sweep"] = mean_sweeps

        if "recorded_thresholds_noise_std" in group:
            noise_stds = _get_group_items(group["recorded_thresholds_noise_std"])
            context_dict["recorded_thresholds_noise_std"] = noise_stds

        if "bg_noise_std" in group:
            bg_noise_std = _get_group_items(group["bg_noise_std"])
            context_dict["bg_noise_std"] = bg_noise_std

        if "sensor_calibration" in group:
            context_dict["sensor_calibration"] = a121.SensorCalibration.from_h5(
                group["sensor_calibration"]
            )

        return SingleSensorContext(**context_dict)


@attrs.mutable(kw_only=True)
class DetectorConfig(AlgoConfigBase):
    start_m: float = attrs.field(default=0.25)
    """Start point of measurement interval in meters."""

    end_m: float = attrs.field(default=3.0)
    """End point of measurement interval in meters."""

    max_step_length: Optional[int] = attrs.field(default=None)  # TODO: Check validity
    """Used to limit step length. If no argument is provided, the step length is automatically
    calculated based on the profile."""

    max_profile: a121.Profile = attrs.field(default=a121.Profile.PROFILE_5, converter=a121.Profile)
    """Specifies the longest allowed profile. If no argument is provided, the highest possible
    profile without interference of direct leakage is used to maximize SNR."""

    signal_quality: float = attrs.field(default=15.0)
    """Signal quality. High quality equals higher HWAAS and better SNR but increase power
    consumption."""

    threshold_method: ThresholdMethod = attrs.field(
        default=ThresholdMethod.CFAR,
        converter=ThresholdMethod,
    )
    """Threshold method"""

    peaksorting_method: PeakSortingMethod = attrs.field(
        default=PeakSortingMethod.HIGHEST_RCS,
        converter=PeakSortingMethod,
    )
    """Sorting method of estimated distances."""

    num_frames_in_recorded_threshold: int = attrs.field(default=100)
    """Number of frames used when calibrating threshold."""

    fixed_threshold_value: float = attrs.field(default=DEFAULT_FIXED_THRESHOLD_VALUE)
    """Value of fixed threshold."""

    threshold_sensitivity: float = attrs.field(default=DEFAULT_THRESHOLD_SENSITIVITY)
    """Sensitivity of threshold. High sensitivity equals low detection threshold, low sensitivity
    equals high detection threshold."""

    update_rate: Optional[float] = attrs.field(default=50.0)
    """Sets the detector update rate."""

    def _collect_validation_results(self) -> list[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []

        if self.end_m < self.start_m:
            validation_results.append(
                a121.ValidationError(
                    self,
                    "start_m",
                    "Must be smaller than 'Range end'",
                )
            )

            validation_results.append(
                a121.ValidationError(
                    self,
                    "end_m",
                    "Must be greater than 'Range start'",
                )
            )

        return validation_results


@attrs.frozen(kw_only=True)
class DetectorResult:
    rcs: Optional[npt.NDArray[np.float_]] = attrs.field(default=None)
    distances: Optional[npt.NDArray[np.float_]] = attrs.field(default=None)
    processor_results: list[ProcessorResult] = attrs.field()
    service_extended_result: list[dict[int, a121.Result]] = attrs.field()


class Detector:
    """Distance detector
    :param client: Client
    :param sensor_id: Sensor id
    :param detector_config: Detector configuration
    :param context: Detector context
    """

    MIN_DIST_M = 0.0
    MAX_DIST_M = 23.0
    MIN_LEAKAGE_FREE_DIST_M = {
        a121.Profile.PROFILE_1: 0.12,
        a121.Profile.PROFILE_2: 0.28,
        a121.Profile.PROFILE_3: 0.56,
        a121.Profile.PROFILE_4: 0.76,
        a121.Profile.PROFILE_5: 1.28,
    }
    MIN_NUM_POINTS_IN_ENVELOPE_FWHM_SPAN = 4.0
    VALID_STEP_LENGTHS = [1, 2, 3, 4, 6, 8, 12, 24]
    NUM_SUBSWEEPS_IN_SENSOR_CONFIG = 4

    MAX_HWAAS = 511
    MIN_HWAAS = 1
    HWAAS_MIN_DISTANCE = 1.0

    session_config: a121.SessionConfig
    processor_specs: List[ProcessorSpec]
    context: DetectorContext

    def __init__(
        self,
        *,
        client: a121.ClientBase,
        sensor_ids: list[int],
        detector_config: DetectorConfig,
        context: Optional[DetectorContext] = None,
    ) -> None:
        self.client = client
        self.sensor_ids = sensor_ids
        self.detector_config = detector_config
        self.started = False

        if context is None or not bool(context.single_sensor_contexts):
            self.context = DetectorContext(
                single_sensor_contexts={
                    sensor_id: SingleSensorContext() for sensor_id in self.sensor_ids
                }
            )
        else:
            self.context = context

        self.aggregator: Optional[Aggregator] = None

        self.update_config(self.detector_config)

    def _validate_ready_for_calibration(self) -> None:
        if self.started:
            raise RuntimeError("Already started")
        if self.processor_specs is None:
            raise ValueError("Processor specification not defined")
        if self.session_config is None:
            raise ValueError("Session config not defined")

    def calibrate_detector(self) -> None:
        """Run the required detector calibration routines, based on the detector config."""

        self._validate_ready_for_calibration()

        self._calibrate_offset()

        self._calibrate_noise()

        if self._has_close_range_measurement(self.detector_config):
            self._calibrate_close_range()

        if self._has_close_range_measurement(
            self.detector_config
        ) or self._has_recorded_threshold_mode(self.detector_config, self.sensor_ids):
            self._record_threshold()

        for context in self.context.single_sensor_contexts.values():
            context.session_config_used_during_calibration = self.session_config

    def _calibrate_close_range(self) -> None:
        """Calibrates the close range measurement parameters used when subtracting the direct
        leakage from the measured signal.

        The parameters calibrated are the direct leakage and a phase reference, used to reduce
        the amount of phase jitter, with the purpose of reducing the residual.
        """

        close_range_spec = self._filter_close_range_spec(self.processor_specs)
        spec = self._update_processor_mode(close_range_spec, ProcessorMode.LEAKAGE_CALIBRATION)

        # Note - Setup with full session_config to match the structure of spec
        extended_metadata = self.client.setup_session(self.session_config)
        assert isinstance(extended_metadata, list)

        aggregators = {
            sensor_id: Aggregator(
                session_config=self.session_config,
                extended_metadata=extended_metadata,
                config=AggregatorConfig(),
                specs=spec,
                sensor_id=sensor_id,
            )
            for sensor_id in self.sensor_ids
        }

        self.client.start_session()
        extended_result = self.client.get_next()
        assert isinstance(extended_result, list)
        self.client.stop_session()

        for sensor_id, context in self.context.single_sensor_contexts.items():
            aggregator_result = aggregators[sensor_id].process(extended_result=extended_result)
            (processor_result,) = aggregator_result.processor_results
            assert processor_result.phase_jitter_comp_reference is not None

            context.direct_leakage = processor_result.direct_leakage
            context.phase_jitter_comp_reference = processor_result.phase_jitter_comp_reference
            context.recorded_thresholds_mean_sweep = None
            context.recorded_thresholds_noise_std = None

            context.sensor_calibration = self.client.calibrations[sensor_id]

    def _record_threshold(self) -> None:
        """Calibrates the parameters used when forming the recorded threshold."""

        # TODO: Ignore/override threshold method while recording threshold

        specs_updated = self._update_processor_mode(
            self.processor_specs, ProcessorMode.RECORDED_THRESHOLD_CALIBRATION
        )

        specs = self._add_context_to_processor_spec(specs_updated)

        extended_metadata = self.client.setup_session(self.session_config)
        assert isinstance(extended_metadata, list)
        aggregators = {
            sensor_id: Aggregator(
                session_config=self.session_config,
                extended_metadata=extended_metadata,
                config=AggregatorConfig(),
                specs=specs[sensor_id],
                sensor_id=sensor_id,
            )
            for sensor_id in self.sensor_ids
        }

        self.client.start_session()
        aggregators_result = {}
        for _ in range(self.detector_config.num_frames_in_recorded_threshold):
            extended_result = self.client.get_next()
            assert isinstance(extended_result, list)
            aggregators_result = {
                sensor_id: aggregators[sensor_id].process(extended_result=extended_result)
                for sensor_id in self.sensor_ids
            }
        self.client.stop_session()

        assert isinstance(extended_result, list)

        for sensor_id, context in self.context.single_sensor_contexts.items():
            recorded_thresholds_mean_sweep = []
            recorded_thresholds_noise_std = []
            for processor_result in aggregators_result[sensor_id].processor_results:
                # Since we know what mode the processor is running in
                assert processor_result.recorded_threshold_mean_sweep is not None
                assert processor_result.recorded_threshold_noise_std is not None

                recorded_thresholds_mean_sweep.append(
                    processor_result.recorded_threshold_mean_sweep
                )
                recorded_thresholds_noise_std.append(processor_result.recorded_threshold_noise_std)

            context.recorded_thresholds_mean_sweep = recorded_thresholds_mean_sweep
            context.recorded_thresholds_noise_std = recorded_thresholds_noise_std
            # Grab temperature from first group as it is the same for all.
            context.reference_temperature = extended_result[0][sensor_id].temperature

    def _calibrate_noise(self) -> None:
        """Estimates the standard deviation of the noise in each subsweep by setting enable_tx to
        False and collecting data, used to calculate the deviation.

        The calibration procedure can be done at any time as it is performed with Tx off, and is
        not effected by objects in front of the sensor.

        This function is called from the start() in the case of CFAR and Fixed threshold and from
        record_threshold() in the case of Recorded threshold. The reason for calling from
        record_threshold() is that it is used when calculating the threshold.
        """

        session_config = copy.deepcopy(self.session_config)

        for sensor_id in self.sensor_ids:
            for group in session_config.groups:
                group[sensor_id].sweeps_per_frame = 1
                for subsweep in group[sensor_id].subsweeps:
                    subsweep.enable_tx = False
                    subsweep.step_length = 1
                    subsweep.start_point = 0
                    # Set num_points to a high number to get sufficient number of data points to
                    # estimate the standard deviation.
                    subsweep.num_points = 500

        extended_metadata = self.client.setup_session(session_config)
        assert isinstance(extended_metadata, list)

        self.client.start_session()
        extended_result = self.client.get_next()
        assert isinstance(extended_result, list)
        self.client.stop_session()

        for sensor_id, context in self.context.single_sensor_contexts.items():
            bg_noise_one_sensor = []
            for spec in self.processor_specs:
                result = extended_result[spec.group_index][sensor_id]
                sensor_config = self.session_config.groups[spec.group_index][sensor_id]
                subsweep_configs = sensor_config.subsweeps
                bg_noise_std_in_subsweep = []
                for idx in spec.subsweep_indexes:
                    if not subsweep_configs[idx].enable_loopback:
                        subframe = result.subframes[idx]
                        subsweep_std = calculate_bg_noise_std(subframe, subsweep_configs[idx])
                        bg_noise_std_in_subsweep.append(subsweep_std)
                bg_noise_one_sensor.append(bg_noise_std_in_subsweep)
            context.bg_noise_std = bg_noise_one_sensor

    def _calibrate_offset(self) -> None:
        """Estimates sensor offset error based on loopback measurement."""

        self._validate_ready_for_calibration()

        sensor_config = a121.SensorConfig(
            start_point=-30,
            num_points=50,
            step_length=1,
            profile=a121.Profile.PROFILE_1,
            hwaas=64,
            sweeps_per_frame=1,
            enable_loopback=True,
            phase_enhancement=True,
        )

        session_config = a121.SessionConfig(
            {sensor_id: sensor_config for sensor_id in self.sensor_ids}, extended=True
        )
        self.client.setup_session(session_config)
        self.client.start_session()
        extended_result = self.client.get_next()
        self.client.stop_session()

        assert isinstance(extended_result, list)

        for sensor_id, context in self.context.single_sensor_contexts.items():
            context.loopback_peak_location_m = calculate_loopback_peak_location(
                extended_result[0][sensor_id], sensor_config
            )

    @staticmethod
    def _get_sensor_calibrations(context: DetectorContext) -> dict[int, a121.SensorCalibration]:
        return {
            sensor_id: single_context.sensor_calibration
            for sensor_id, single_context in context.single_sensor_contexts.items()
            if single_context.sensor_calibration is not None
        }

    @classmethod
    def get_detector_status(
        cls,
        config: DetectorConfig,
        context: DetectorContext,
        sensor_ids: list[int],
    ) -> DetectorStatus:
        """Returns the detector status along with the detector state."""

        if not cls._valid_detector_config_range(config=config):
            return DetectorStatus(
                detector_state=DetailedStatus.INVALID_DETECTOR_CONFIG_RANGE,
                ready_to_start=False,
            )

        if config.end_m < config.start_m:
            return DetectorStatus(
                detector_state=DetailedStatus.END_LESSER_THAN_START,
                ready_to_start=False,
            )

        if len(sensor_ids) != len(set(sensor_ids)):
            return DetectorStatus(
                detector_state=DetailedStatus.SENSOR_IDS_NOT_UNIQUE,
                ready_to_start=False,
            )

        if context.single_sensor_contexts is None:
            return DetectorStatus(
                detector_state=DetailedStatus.CONTEXT_MISSING,
                ready_to_start=False,
            )

        (session_config, _,) = cls._detector_to_session_config_and_processor_specs(
            config=config, sensor_ids=sensor_ids
        )

        # Offset calibration is always performed as a part of the detector calibration process.
        # Use this as indication whether detector calibration has been performed.
        calibration_missing = np.any(
            [
                ctx.loopback_peak_location_m is None
                for ctx in context.single_sensor_contexts.values()
            ]
        )
        config_mismatch = np.any(
            [
                ctx.session_config_used_during_calibration != session_config
                for ctx in context.single_sensor_contexts.values()
            ]
        )

        if calibration_missing:
            detector_state = DetailedStatus.CALIBRATION_MISSING
        elif config_mismatch:
            detector_state = DetailedStatus.CONFIG_MISMATCH
        else:
            detector_state = DetailedStatus.OK

        return DetectorStatus(
            detector_state=detector_state,
            ready_to_start=(detector_state == DetailedStatus.OK),
        )

    @classmethod
    def _valid_detector_config_range(cls, config: DetectorConfig) -> bool:
        return cls.MIN_DIST_M <= config.start_m and config.end_m <= cls.MAX_DIST_M

    @staticmethod
    def _close_range_calibrated(context: DetectorContext) -> bool:
        has_dl = np.all(
            [ctx.direct_leakage is not None for ctx in context.single_sensor_contexts.values()]
        )
        has_pjcr = np.all(
            [
                ctx.phase_jitter_comp_reference is not None
                for ctx in context.single_sensor_contexts.values()
            ]
        )

        if has_dl != has_pjcr:
            raise RuntimeError

        return bool(has_dl and has_pjcr)

    @staticmethod
    def _recorded_threshold_calibrated(context: DetectorContext) -> bool:
        mean_sweep_calibrated = np.all(
            [
                ctx.recorded_thresholds_mean_sweep is not None
                for ctx in context.single_sensor_contexts.values()
            ]
        )

        std_calibrated = np.all(
            [
                ctx.recorded_thresholds_noise_std is not None
                for ctx in context.single_sensor_contexts.values()
            ]
        )

        return bool(mean_sweep_calibrated and std_calibrated)

    @classmethod
    def _has_close_range_measurement(self, config: DetectorConfig) -> bool:
        # sensor_ids=[1] as the detector is running the same config for all sensors.
        (
            _,
            specs,
        ) = self._detector_to_session_config_and_processor_specs(config=config, sensor_ids=[1])
        return MeasurementType.CLOSE_RANGE in [
            spec.processor_config.measurement_type for spec in specs
        ]

    @classmethod
    def _has_recorded_threshold_mode(self, config: DetectorConfig, sensor_ids: list[int]) -> bool:
        (_, processor_specs,) = self._detector_to_session_config_and_processor_specs(
            config=config, sensor_ids=sensor_ids
        )
        return ThresholdMethod.RECORDED in [
            spec.processor_config.threshold_method for spec in processor_specs
        ]

    def start(
        self,
        recorder: Optional[a121.Recorder] = None,
        *,
        _algo_group: Optional[h5py.Group] = None,
    ) -> None:
        """Method for setting up measurement session."""

        if self.started:
            raise RuntimeError("Already started")

        status = self.get_detector_status(self.detector_config, self.context, self.sensor_ids)

        if not status.ready_to_start:
            raise RuntimeError(f"Not ready to start ({status.detector_state.name})")
        specs = self._add_context_to_processor_spec(self.processor_specs)

        sensor_calibration = self._get_sensor_calibrations(self.context)

        extended_metadata = self.client.setup_session(
            self.session_config, calibrations=sensor_calibration
        )

        assert isinstance(extended_metadata, list)
        assert np.all(
            [
                context.loopback_peak_location_m is not None
                for context in self.context.single_sensor_contexts.values()
            ]
        )
        aggregator_config = AggregatorConfig(
            peak_sorting_method=self.detector_config.peaksorting_method
        )
        self.aggregators = {
            sensor_id: Aggregator(
                session_config=self.session_config,
                extended_metadata=extended_metadata,
                config=aggregator_config,
                specs=specs[sensor_id],
                sensor_id=sensor_id,
            )
            for sensor_id in self.sensor_ids
        }

        if recorder is not None:
            if isinstance(recorder, a121.H5Recorder):
                if _algo_group is None:
                    _algo_group = recorder.require_algo_group("distance_detector")

                _record_algo_data(
                    _algo_group,
                    self.sensor_ids,
                    self.detector_config,
                    self.context,
                )
            else:
                # Should never happen as we currently only have the H5Recorder
                warnings.warn("Will not save algo data")

        self.client.start_session(recorder)
        self.started = True

    def get_next(self) -> Dict[int, DetectorResult]:
        """Called from host to get next measurement."""
        if not self.started:
            raise RuntimeError("Not started")

        assert self.aggregators is not None

        extended_result = self.client.get_next()
        assert isinstance(extended_result, list)

        aggregator_results = {
            sensor_id: self.aggregators[sensor_id].process(extended_result=extended_result)
            for sensor_id in self.sensor_ids
        }

        result = {
            sensor_id: DetectorResult(
                rcs=aggregator_results[sensor_id].estimated_rcs,
                distances=aggregator_results[sensor_id].estimated_distances,
                processor_results=aggregator_results[sensor_id].processor_results,
                service_extended_result=aggregator_results[sensor_id].service_extended_result,
            )
            for sensor_id in self.sensor_ids
        }

        return result

    def update_config(self, config: DetectorConfig) -> None:
        """Updates the session config and processor specification based on the detector
        configuration."""
        (
            self.session_config,
            self.processor_specs,
        ) = self._detector_to_session_config_and_processor_specs(
            config=config, sensor_ids=self.sensor_ids
        )

    def stop(self) -> Any:
        """Stops the measurement session."""
        if not self.started:
            raise RuntimeError("Already stopped")

        recorder_result = self.client.stop_session()

        self.started = False

        return recorder_result

    @classmethod
    def _detector_to_session_config_and_processor_specs(
        cls, config: DetectorConfig, sensor_ids: list[int]
    ) -> Tuple[a121.SessionConfig, list[ProcessorSpec]]:
        processor_specs = []
        groups = []
        group_index = 0

        plans = cls._create_group_plans(config)

        if MeasurementType.CLOSE_RANGE in plans:
            sensor_config = cls._close_subsweep_group_plans_to_sensor_config(
                plans[MeasurementType.CLOSE_RANGE]
            )
            groups.append({sensor_id: sensor_config for sensor_id in sensor_ids})
            processor_specs.append(
                ProcessorSpec(
                    processor_config=ProcessorConfig(
                        threshold_method=ThresholdMethod.RECORDED,
                        measurement_type=MeasurementType.CLOSE_RANGE,
                        threshold_sensitivity=config.threshold_sensitivity,
                    ),
                    group_index=group_index,
                    subsweep_indexes=[0, 1],
                )
            )
            group_index += 1

        if MeasurementType.FAR_RANGE in plans:
            (
                sensor_config,
                processor_specs_subsweep_indexes,
            ) = cls._far_subsweep_group_plans_to_sensor_config_and_subsweep_indexes(
                plans[MeasurementType.FAR_RANGE]
            )
            groups.append({sensor_id: sensor_config for sensor_id in sensor_ids})

            processor_config = ProcessorConfig(
                threshold_method=config.threshold_method,
                fixed_threshold_value=config.fixed_threshold_value,
                threshold_sensitivity=config.threshold_sensitivity,
            )

            for subsweep_indexes in processor_specs_subsweep_indexes:
                processor_specs.append(
                    ProcessorSpec(
                        processor_config=processor_config,
                        group_index=group_index,
                        subsweep_indexes=subsweep_indexes,
                    )
                )

        return (
            a121.SessionConfig(groups, extended=True, update_rate=config.update_rate),
            processor_specs,
        )

    @classmethod
    def _create_group_plans(
        cls, config: DetectorConfig
    ) -> Dict[MeasurementType, List[SubsweepGroupPlan]]:
        """
        Create dictionary containing group plans for close and far range measurements.

        - Close range measurement: Add Subsweep group if the user defined starting point is
        effected by the direct leakage.
        - Transition region: Add group plans to bridge the gap between the start of the far range
        measurement region(either end of close range region or user defined start_m) and the
        shortest measurable distance with max_profile, free from direct leakage interference.
        - Add group plan with max_profile. Increase HWAAS as a function of distance to maintain
        SNR throughout the sweep.
        """
        plans = {}

        # Determine shortest direct leakage free distance per profile
        min_dist_m = cls._calc_leakage_free_min_dist(config)

        close_range_transition_m = min_dist_m[a121.Profile.PROFILE_1]

        # Add close range group plan if applicable
        if config.start_m < close_range_transition_m:
            plans[MeasurementType.CLOSE_RANGE] = cls._get_close_range_group_plan(
                close_range_transition_m, config
            )

        # Define transition group plans
        transition_subgroup_plans = cls._get_transition_group_plans(
            config, min_dist_m, MeasurementType.CLOSE_RANGE in plans
        )

        # The number of available subsweeps in the group with max profile.
        num_remaining_subsweeps = cls.NUM_SUBSWEEPS_IN_SENSOR_CONFIG - len(
            transition_subgroup_plans
        )

        # No neighbours if no close range measurement or transition groups defined.
        has_neighbouring_subsweep = (
            MeasurementType.CLOSE_RANGE in plans or len(transition_subgroup_plans) != 0
        )

        # Define group plans with max profile
        max_profile_subgroup_plans = cls._get_max_profile_group_plans(
            config, min_dist_m, has_neighbouring_subsweep, num_remaining_subsweeps
        )

        far_subgroup_plans = transition_subgroup_plans + max_profile_subgroup_plans

        if len(far_subgroup_plans) != 0:
            plans[MeasurementType.FAR_RANGE] = far_subgroup_plans

        return plans

    @classmethod
    def _get_close_range_group_plan(
        cls, transition_m: float, config: DetectorConfig
    ) -> list[SubsweepGroupPlan]:
        """Define the group plan for close range measurements.

        The close range measurement always use profile 1 to minimize direct leakage region.
        """
        profile = a121.Profile.PROFILE_1
        # No left neighbour as this is the first subsweep when close range measurement is
        # applicable.
        has_neighbour = (False, transition_m < config.end_m)
        return [
            cls._create_group_plan(
                profile, config, [config.start_m, transition_m], has_neighbour, True
            )
        ]

    @classmethod
    def _get_transition_group_plans(
        cls,
        config: DetectorConfig,
        min_dist_m: Dict[a121.Profile, float],
        has_close_range_measurement: bool,
    ) -> list[SubsweepGroupPlan]:
        """Define the transition segment group plans.

        The purpose of the transition group is to bridge the gap between the start point of the
        far measurement region and the point where max_profile can be used without interference
        of direct leakage.

        The transition region can consist of maximum two subsweeps, where the first utilize profile
        1 and the second profile 3. Whether both, one or none is used depends on the user provided
        detector config.
        """
        transition_profiles = [
            profile
            for profile in [a121.Profile.PROFILE_1, a121.Profile.PROFILE_3]
            if profile.value < config.max_profile.value
        ]
        transition_profiles.append(config.max_profile)

        transition_subgroup_plans: list = []

        for i in range(len(transition_profiles) - 1):
            profile = transition_profiles[i]
            next_profile = transition_profiles[i + 1]

            if config.start_m < min_dist_m[next_profile] and min_dist_m[profile] < config.end_m:
                start_m = max(min_dist_m[profile], config.start_m)
                end_m = min(config.end_m, min_dist_m[next_profile])
                has_neighbour = (
                    has_close_range_measurement or len(transition_subgroup_plans) != 0,
                    min_dist_m[next_profile] < end_m,
                )

                transition_subgroup_plans.append(
                    cls._create_group_plan(profile, config, [start_m, end_m], has_neighbour, False)
                )

        return transition_subgroup_plans

    @classmethod
    def _get_max_profile_group_plans(
        cls,
        config: DetectorConfig,
        min_dist_m: Dict[a121.Profile, float],
        has_neighbouring_subsweep: bool,
        num_remaining_subsweeps: int,
    ) -> list[SubsweepGroupPlan]:
        """Define far range group plans with max_profile

        Divide the measurement range from the shortest leakage free distance of max_profile to
        the end point into equidistance segments and assign HWAAS according to the radar equation
        to maintain SNR throughout the sweep.
        """

        if min_dist_m[config.max_profile] < config.end_m:
            subsweep_start_m = max([config.start_m, min_dist_m[config.max_profile]])
            breakpoints_m = np.linspace(
                subsweep_start_m,
                config.end_m,
                num_remaining_subsweeps + 1,
            ).tolist()

            return [
                cls._create_group_plan(
                    config.max_profile,
                    config,
                    breakpoints_m,
                    (has_neighbouring_subsweep, False),
                    False,
                )
            ]
        else:
            return []

    @classmethod
    def _create_group_plan(
        cls,
        profile: a121.Profile,
        config: DetectorConfig,
        breakpoints_m: list[float],
        has_neighbour: Tuple[bool, bool],
        is_close_range_measurement: bool,
    ) -> SubsweepGroupPlan:
        """Creates a group plan."""
        step_length = cls._limit_step_length(profile, config.max_step_length)
        breakpoints = cls._m_to_points(breakpoints_m, step_length)
        hwaas = cls._calculate_hwaas(profile, breakpoints, config.signal_quality, step_length)

        extended_breakpoints = cls._add_margin_to_breakpoints(
            profile=profile,
            step_length=step_length,
            base_bpts=breakpoints,
            has_neighbour=has_neighbour,
            config=config,
            is_close_range_measurement=is_close_range_measurement,
        )

        return SubsweepGroupPlan(
            step_length=step_length,
            breakpoints=extended_breakpoints,
            profile=profile,
            hwaas=hwaas,
        )

    @classmethod
    def _calc_leakage_free_min_dist(cls, config: DetectorConfig) -> Dict[a121.Profile, float]:
        """This function calculates the shortest leakage free distance per profile, for all profiles
        up to max_profile"""
        min_dist_m = {}
        for profile, min_dist in cls.MIN_LEAKAGE_FREE_DIST_M.items():
            min_dist_m[profile] = min_dist
            if config.threshold_method == ThresholdMethod.CFAR:
                step_length = cls._limit_step_length(profile, config.max_step_length)
                cfar_margin_m = (
                    Processor.calc_cfar_margin(profile, step_length)
                    * step_length
                    * APPROX_BASE_STEP_LENGTH_M
                )
                min_dist_m[profile] += cfar_margin_m

            if profile == config.max_profile:
                # All profiles up to max_profile has been added. Break and return result.
                break

        return min_dist_m

    @classmethod
    def _calculate_hwaas(
        cls, profile: a121.Profile, breakpoints: list[int], signal_quality: float, step_length: int
    ) -> list[int]:
        rlg_per_hwaas = Processor.RLG_PER_HWAAS_MAP[profile]
        hwaas = []
        for idx in range(len(breakpoints) - 1):
            processing_gain = Processor.calc_processing_gain(profile, step_length)
            subsweep_end_point_m = max(
                APPROX_BASE_STEP_LENGTH_M * breakpoints[idx + 1],
                cls.HWAAS_MIN_DISTANCE,
            )
            rlg = signal_quality + 40 * np.log10(subsweep_end_point_m) - np.log10(processing_gain)
            hwaas_in_subsweep = int(10 ** ((rlg - rlg_per_hwaas) / 10))
            hwaas.append(np.clip(hwaas_in_subsweep, cls.MIN_HWAAS, cls.MAX_HWAAS))
        return hwaas

    @classmethod
    def _add_margin_to_breakpoints(
        cls,
        profile: a121.Profile,
        step_length: int,
        base_bpts: list[int],
        has_neighbour: Tuple[bool, bool],
        config: DetectorConfig,
        is_close_range_measurement: bool,
    ) -> list[int]:
        """
        Add points to segment edges based on their position.

        1. Add one margin to each segment for distance filter initialization
        2. Add an additional margin to segments with neighbouring segments for segment overlap
        """

        margin_p = get_distance_filter_edge_margin(profile, step_length) * step_length
        left_margin = margin_p
        right_margin = margin_p

        if has_neighbour[0]:
            left_margin += margin_p

        if has_neighbour[1]:
            right_margin += margin_p

        if config.threshold_method == ThresholdMethod.CFAR and not is_close_range_measurement:
            cfar_margin = Processor.calc_cfar_margin(profile, step_length) * step_length
            left_margin += cfar_margin
            right_margin += cfar_margin

        bpts = copy.copy(base_bpts)
        bpts[0] -= left_margin
        bpts[-1] += right_margin

        return bpts

    @classmethod
    def _limit_step_length(cls, profile: a121.Profile, user_limit: Optional[int]) -> int:
        """
        Calculates step length based on user defined step length and selected profile.

        The step length must yield minimum MIN_NUM_POINTS_IN_ENVELOPE_FWHM_SPAN number of points
        in the span of the FWHM of the envelope.

        If the step length is <24, return the valid step length(defined by
        VALID_STEP_LENGTHS) that is closest to, but not longer than the limit.

        If the limit is 24<=, return the multiple of 24 that is
        closest, but not longer than the limit.
        """

        fwhm_p = ENVELOPE_FWHM_M[profile] / APPROX_BASE_STEP_LENGTH_M
        limit = int(fwhm_p / cls.MIN_NUM_POINTS_IN_ENVELOPE_FWHM_SPAN)

        if user_limit is not None:
            limit = min(user_limit, limit)

        if limit < cls.VALID_STEP_LENGTHS[-1]:
            idx_closest = np.sum(np.array(cls.VALID_STEP_LENGTHS) <= limit) - 1
            return int(cls.VALID_STEP_LENGTHS[idx_closest])
        else:
            return int((limit // cls.VALID_STEP_LENGTHS[-1]) * cls.VALID_STEP_LENGTHS[-1])

    @classmethod
    def _close_subsweep_group_plans_to_sensor_config(
        cls, plan_: List[SubsweepGroupPlan]
    ) -> a121.SensorConfig:
        (plan,) = plan_
        subsweeps = []
        subsweeps.append(
            a121.SubsweepConfig(
                start_point=0,
                num_points=1,
                step_length=1,
                profile=a121.Profile.PROFILE_4,
                hwaas=plan.hwaas[0],
                receiver_gain=15,
                phase_enhancement=True,
                enable_loopback=True,
            )
        )
        num_points = int((plan.breakpoints[1] - plan.breakpoints[0]) / plan.step_length)
        subsweeps.append(
            a121.SubsweepConfig(
                start_point=plan.breakpoints[0],
                num_points=num_points,
                step_length=plan.step_length,
                profile=plan.profile,
                hwaas=plan.hwaas[0],
                receiver_gain=5,
                phase_enhancement=True,
                prf=select_prf(plan.breakpoints[1], plan.profile),
            )
        )
        return a121.SensorConfig(subsweeps=subsweeps, sweeps_per_frame=10)

    @classmethod
    def _far_subsweep_group_plans_to_sensor_config_and_subsweep_indexes(
        cls, subsweep_group_plans: list[SubsweepGroupPlan]
    ) -> Tuple[a121.SensorConfig, list[list[int]]]:
        subsweeps = []
        processor_specs_subsweep_indexes = []
        subsweep_idx = 0
        for plan in subsweep_group_plans:
            subsweep_indexes = []
            for bp_idx in range(len(plan.breakpoints) - 1):
                num_points = int(
                    (plan.breakpoints[bp_idx + 1] - plan.breakpoints[bp_idx]) / plan.step_length
                )
                subsweeps.append(
                    a121.SubsweepConfig(
                        start_point=plan.breakpoints[bp_idx],
                        num_points=num_points,
                        step_length=plan.step_length,
                        profile=plan.profile,
                        hwaas=plan.hwaas[bp_idx],
                        receiver_gain=10,
                        phase_enhancement=True,
                        prf=select_prf(plan.breakpoints[bp_idx + 1], plan.profile),
                    )
                )
                subsweep_indexes.append(subsweep_idx)
                subsweep_idx += 1
            processor_specs_subsweep_indexes.append(subsweep_indexes)
        return (
            a121.SensorConfig(subsweeps=subsweeps, sweeps_per_frame=1),
            processor_specs_subsweep_indexes,
        )

    @classmethod
    def _m_to_points(cls, breakpoints_m: list[float], step_length: int) -> list[int]:
        bpts_m = np.array(breakpoints_m)
        start_point = int(bpts_m[0] / APPROX_BASE_STEP_LENGTH_M)
        num_steps = (bpts_m[-1] - bpts_m[0]) / (APPROX_BASE_STEP_LENGTH_M)
        bpts = num_steps / (bpts_m[-1] - bpts_m[0]) * (bpts_m - bpts_m[0]) + start_point
        return [(bpt // step_length) * step_length for bpt in bpts]

    @classmethod
    def _update_processor_mode(
        cls, processor_specs: list[ProcessorSpec], processor_mode: ProcessorMode
    ) -> list[ProcessorSpec]:
        updated_specs = []
        for spec in processor_specs:
            new_processor_config = attrs.evolve(
                spec.processor_config, processor_mode=processor_mode
            )
            updated_specs.append(attrs.evolve(spec, processor_config=new_processor_config))
        return updated_specs

    @classmethod
    def _filter_close_range_spec(cls, specs: list[ProcessorSpec]) -> list[ProcessorSpec]:
        NUM_CLOSE_RANGE_SPECS = 1
        close_range_specs = []
        for spec in specs:
            if spec.processor_config.measurement_type == MeasurementType.CLOSE_RANGE:
                close_range_specs.append(spec)
        if len(close_range_specs) != NUM_CLOSE_RANGE_SPECS:
            raise ValueError("Incorrect subsweep config for close range measurement")

        return close_range_specs

    def _add_context_to_processor_spec(
        self, processor_specs: list[ProcessorSpec]
    ) -> dict[int, list[ProcessorSpec]]:
        """
        Create and add processor context to processor specification.
        """

        updated_specs_all_sensors = {}
        for sensor_id, context in self.context.single_sensor_contexts.items():
            assert context.bg_noise_std is not None
            updated_specs: List[ProcessorSpec] = []
            for idx, (spec, bg_noise_std) in enumerate(zip(processor_specs, context.bg_noise_std)):
                if (
                    context.recorded_thresholds_mean_sweep is not None
                    or context.recorded_thresholds_noise_std is not None
                ):
                    assert context.recorded_thresholds_mean_sweep is not None
                    assert context.recorded_thresholds_noise_std is not None
                    recorded_thresholds_mean_sweep = context.recorded_thresholds_mean_sweep[idx]
                    recorded_threshold_noise_std = context.recorded_thresholds_noise_std[idx]
                else:
                    recorded_thresholds_mean_sweep = None
                    recorded_threshold_noise_std = None

                if (
                    context.direct_leakage is not None
                    or context.phase_jitter_comp_reference is not None
                ):
                    direct_leakage = context.direct_leakage
                    phase_jitter_comp_ref = context.phase_jitter_comp_reference
                else:
                    direct_leakage = None
                    phase_jitter_comp_ref = None

                if context.loopback_peak_location_m is not None:
                    loopback_peak_location_m = context.loopback_peak_location_m
                else:
                    loopback_peak_location_m = None

                if context.reference_temperature is not None:
                    reference_temperature = context.reference_temperature
                else:
                    reference_temperature = None

                processor_context = ProcessorContext(
                    recorded_threshold_mean_sweep=recorded_thresholds_mean_sweep,
                    recorded_threshold_noise_std=recorded_threshold_noise_std,
                    bg_noise_std=bg_noise_std,
                    direct_leakage=direct_leakage,
                    phase_jitter_comp_ref=phase_jitter_comp_ref,
                    reference_temperature=reference_temperature,
                    loopback_peak_location_m=loopback_peak_location_m,
                )
                updated_specs.append(attrs.evolve(spec, processor_context=processor_context))
            updated_specs_all_sensors[sensor_id] = updated_specs
        return updated_specs_all_sensors


def _record_algo_data(
    algo_group: h5py.Group,
    sensor_ids: list[int],
    detector_config: DetectorConfig,
    context: DetectorContext,
) -> None:
    algo_group.create_dataset(
        "sensor_ids",
        data=sensor_ids,
        track_times=False,
    )
    _create_h5_string_dataset(algo_group, "detector_config", detector_config.to_json())

    context_group = algo_group.create_group("context")
    context.to_h5(context_group)


def _load_algo_data(
    algo_group: h5py.Group,
) -> Tuple[int, DetectorConfig, DetectorContext]:
    sensor_id = algo_group["sensor_ids"][()]
    config = DetectorConfig.from_json(algo_group["detector_config"][()])

    context_group = algo_group["context"]
    context = DetectorContext.from_h5(context_group)

    return sensor_id, config, context


def _get_group_items(group: h5py.Group) -> list[npt.NDArray]:
    group_items = []

    i = 0
    while True:
        try:
            v = group[f"index_{i}"][()]
        except KeyError:
            break

        group_items.append(v)
        i += 1
    return group_items



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from typing import Optional, Tuple

import requests


def check_package_outdated(name: str, current_version: str) -> Tuple[bool, Optional[str]]:

    try:
        response = requests.get(f"https://pypi.python.org/pypi/{name}/json")
        latest_version = response.json()["info"]["version"]
    except Exception:
        latest_version = None
        is_outdated = False
        return is_outdated, latest_version

    if latest_version == current_version:
        is_outdated = False
    else:
        is_outdated = True

    return is_outdated, latest_version


def get_latest_changelog() -> str:
    try:
        cl = requests.get(
            "https://raw.githubusercontent.com/"
            "acconeer/acconeer-python-exploration/master/CHANGELOG.md"
        )
        return str(cl.content.decode())
    except Exception:
        return str("### No changelog available at this time")



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import argparse

import numpy as np

from acconeer.exptool import a121


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("filename")
    parser.add_argument("--plot", action="store_true")
    args = parser.parse_args()
    filename = args.filename

    record = a121.load_record(filename)

    print_record(record)

    if args.plot:
        plot_record(record)


def print_record(record: a121.Record) -> None:
    estimated_update_rate = 1 / np.diff(record.stacked_results.tick_time).mean()

    print("ET version: ", record.lib_version)
    print("RSS version:", record.server_info.rss_version)
    print("HW name:    ", record.server_info.hardware_name)
    print("Data shape: ", record.frames.shape)
    print("Est. rate:  ", f"{estimated_update_rate:.3f} Hz")
    print("Timestamp:  ", record.timestamp)
    print()
    print(record.session_config)
    print()
    print(record.metadata)
    print()

    first_result = next(record.results)
    print(first_result)


def plot_record(record: a121.Record) -> None:
    import matplotlib.pyplot as plt

    fig, ax = plt.subplots()

    z = abs(record.frames.mean(axis=1)).T
    x = record.stacked_results.tick_time
    x -= x[0]
    y = np.arange(z.shape[0])

    ax.pcolormesh(x, y, z)
    ax.set_xlabel("Time (s)")
    ax.set_ylabel("Distance point")
    ax.set_title("Mean sweep amplitude")

    fig.tight_layout()
    plt.show()


if __name__ == "__main__":
    main()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import abc

from PySide6.QtWidgets import QVBoxLayout, QWidget

from acconeer.exptool.app.new import PluginState
from acconeer.exptool.app.new.app_model import AppModel, ViewPluginInterface
from acconeer.exptool.app.new.backend import GeneralMessage
from acconeer.exptool.app.new.ui import utils

from .ui_plugin_base import UiPluginBase


class ViewPluginBase(UiPluginBase, abc.ABC, ViewPluginInterface):
    def __init__(self, app_model: AppModel, view_widget: QWidget) -> None:
        super().__init__(app_model)
        self.app_model = app_model
        self.app_model.sig_message_view_plugin.connect(self.handle_message)

        self.__view_widget = view_widget
        self.__view_widget.setLayout(QVBoxLayout())

        self._sticky_widget = QWidget()
        self._scrolly_widget = QWidget()

        self.__view_widget.layout().addWidget(self._sticky_widget)
        self.__view_widget.layout().addWidget(utils.HorizontalSeparator())
        self.__view_widget.layout().addWidget(
            utils.ScrollAreaDecorator(
                utils.TopAlignDecorator(
                    self._scrolly_widget,
                )
            )
        )

    def stop_listening(self) -> None:
        super().stop_listening()
        self.app_model.sig_message_view_plugin.disconnect(self.handle_message)

    @property
    def sticky_widget(self) -> QWidget:
        """The sticky widget. The sticky area is located at the top."""
        return self._sticky_widget

    @property
    def scrolly_widget(self) -> QWidget:
        """The scrolly widget. The scrolly area is located below the sticky area"""
        return self._scrolly_widget

    @abc.abstractmethod
    def handle_message(self, message: GeneralMessage) -> None:
        pass

    def _send_start_request(self) -> None:
        self.app_model.put_backend_plugin_task(
            "start_session",
            {"with_recorder": self.app_model.recording_enabled},
            on_error=self.app_model.emit_error,
        )
        self.app_model.set_plugin_state(PluginState.LOADED_STARTING)

    def _send_stop_request(self) -> None:
        self.app_model.put_backend_plugin_task("stop_session", on_error=self.app_model.emit_error)
        self.app_model.set_plugin_state(PluginState.LOADED_STOPPING)
